from fluid_sbom.internal.cache import (
    dual_cache,
)
import requests
from typing import (
    Any,
)


@dual_cache
def get_nuget_package(
    package_name: str, version: str | None = None
) -> dict[str, Any] | None:
    package_name = package_name.lower()
    base_url = (
        f"https://api.nuget.org/v3/registration5-gz-semver2/{package_name}"
    )
    if version:
        base_url += f"/{version}"
    else:
        base_url += "/index"
    base_url += ".json"
    response = requests.get(base_url, timeout=20)
    if response.status_code == 200:
        package_data = response.json()
        if version:
            package_data = requests.get(
                package_data["catalogEntry"], timeout=20
            ).json()
        else:
            items = requests.get(
                package_data["items"][-1]["@id"], timeout=20
            ).json()["items"]
            try:
                package_data = next(
                    x["catalogEntry"]
                    for x in reversed(items)
                    if "catalogEntry" in x
                )
            except StopIteration:
                package_data = next(
                    y["catalogEntry"]
                    for x in reversed(items)
                    for y in reversed(x.get("items", []))
                    if "catalogEntry" in y
                    # not list preview versions
                    and "pre" not in y["catalogEntry"]["version"]
                )

        return package_data

    return None



from fluid_sbom.pkg.cataloger.generic.cataloger import (
    Request,
)
from fluid_sbom.pkg.cataloger.redhat.parse_rpm_db import (
    parse_rpm_db,
)
from fnmatch import (
    fnmatch,
)
import reactivex
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)


def on_next_redhat(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if any(
                    fnmatch(value, x)
                    for x in (
                        # /var/lib/rpm/Packages
                        # /var/lib/rpm/Packages.db
                        "/var/lib/rpm/rpmdb.sqlite",
                        # /usr/share/rpm/Packages
                        # /usr/share/rpm/Packages.db
                        "/usr/share/rpm/rpmdb.sqlite",
                        # /usr/lib/sysimage/rpm/Packages
                        # /usr/lib/sysimage/rpm/Packages.db
                        "/usr/lib/sysimage/rpm/rpmdb.sqlite",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_rpm_db,
                            parser_name="redhat-parse-rpmdb",
                        )
                    )
                if any(fnmatch(value, x) for x in ("**/*.rpm", "*.rpm")):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_rpm_db,
                            parser_name="redhat-parse-rpmdb",
                        )
                    )
            except Exception as ex:  # pylint:disable=broad-exception-caught
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from fluid_sbom.file.type import (
    Type,
)
from pydantic import (
    BaseModel,
)


class Metadata(BaseModel):
    path: str
    link_destination: str
    user_id: int
    group_id: int
    type: Type
    mime_type: str



from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.generic.cataloger import (
    Request,
    Task,
)
import reactivex
from reactivex import (
    Observable,
)
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)
from typing import (
    Callable,
)


def gen_location(
    resolver: Resolver,
) -> Callable[[Observable[Request]], Observable]:
    def _handle(source: Observable[Request]) -> Observable:
        def subscribe(
            observer: ObserverBase[Task],
            scheduler: SchedulerBase | None = None,
        ) -> reactivex.abc.DisposableBase:
            def on_next(value: Request) -> None:
                try:
                    locations = resolver.files_by_path(value.real_path)
                    for location in locations:
                        observer.on_next(
                            Task(
                                location=location,
                                parser=value.parser,
                                parser_name=value.parser_name,
                            )
                        )
                except (
                    Exception  # pylint:disable=broad-exception-caught
                ) as ex:
                    observer.on_error(ex)

            return source.subscribe(
                on_next,
                observer.on_error,
                observer.on_completed,
                scheduler=scheduler,
            )

        return reactivex.create(subscribe)

    return _handle



from enum import (
    Enum,
)
from tarfile import (
    TarInfo,
)


class Type(Enum):
    TYPE_REGULAR = "TypeRegular"
    TYPE_HARD_LINK = "TypeHardLink"
    TYPE_SYM_LINK = "TypeSymLink"
    TYPE_CHARACTER_DEVICE = "TypeCharacterDevice"
    TYPE_BLOCK_DEVICE = "TypeBlockDevice"
    TYPE_DIRECTORY = "TypeDirectory"
    TYPE_FIFO = "TypeFIFO"
    TYPE_SOCKET = "TypeSocket"
    TYPE_IRREGULAR = "TypeIrregular"


def get_type_from_tar_member(  # pylint:disable=too-many-return-statements
    member: TarInfo,
) -> Type:
    if member.issym():
        return Type.TYPE_SYM_LINK
    if member.islnk():
        return Type.TYPE_HARD_LINK
    if member.isreg():
        return Type.TYPE_REGULAR
    if member.isdir():
        return Type.TYPE_DIRECTORY
    if member.isfifo():
        return Type.TYPE_FIFO
    if member.ischr():
        return Type.TYPE_CHARACTER_DEVICE
    if member.isblk():
        return Type.TYPE_BLOCK_DEVICE
    return Type.TYPE_IRREGULAR



from fluid_sbom.internal.package_information.dart import (
    get_pub_package,
)
from fluid_sbom.pkg.package import (
    Artifact,
    HealthMetadata,
    Package,
)
from fluid_sbom.utils.file import (
    Digest,
)


def complete_package(package: Package) -> Package:
    pub_package = get_pub_package(package.name)
    if not pub_package:
        return package

    current_package = next(
        (
            version
            for version in pub_package["versions"]
            if version["version"] == package.version
        ),
        None,
    )
    package.health_metadata = HealthMetadata(
        latest_version=pub_package["latest"]["version"],
        latest_version_created_at=pub_package["latest"]["published"],
    )
    if current_package:
        package.health_metadata.artifact = Artifact(
            url=current_package["archive_url"],
            integrity=Digest(
                value=current_package["archive_sha256"],
                algorithm="sha256",
            ),
        )

    package.health_metadata.authors = next(
        (
            version["pubspec"]["author"]
            for version in reversed(pub_package["versions"])
            if "author" in version["pubspec"]
        ),
        None,
    )

    return package



from datetime import (
    datetime,
)
from fluid_sbom.utils.file import (
    Digest,
)
from pydantic import (
    BaseModel,
)


class AlpmFileRecord(BaseModel):
    path: str
    type: str | None = None
    uid: str | None = None
    gid: str | None = None
    time: datetime | None = None
    size: str | None = None
    link: str | None = None
    digests: list[Digest] | None = None


class AlpmDBEntry(BaseModel):  # pylint:disable=too-many-instance-attributes
    licenses: str = ""
    base_package: str = ""
    package: str = ""
    version: str = ""
    description: str = ""
    architecture: str = ""
    size: int = 0
    packager: str = ""
    url: str = ""
    validation: str = ""
    reason: int = 0
    files: list[AlpmFileRecord] | None = None
    backup: list[AlpmFileRecord] | None = None

    def owned_files(self) -> list[str]:
        seen = set()
        result = []
        for file_record in self.files or []:
            if file_record.path and file_record.path not in seen:
                seen.add(file_record.path)
                result.append(file_record.path)
        result.sort()
        return result



from pydantic import (
    BaseModel,
)


class PhpComposerExternalReference(BaseModel):
    type: str
    url: str
    reference: str
    shasum: str | None = None


class PhpComposerAuthors(BaseModel):
    name: str
    email: str | None = None
    homepage: str | None = None


class PhpComposerLockEntry(BaseModel):
    name: str
    version: str
    source: PhpComposerExternalReference
    dist: PhpComposerExternalReference
    require: dict[str, str] | None = None
    provide: dict[str, str] | None = None
    require_dev: dict[str, str] | None = None
    suggest: dict[str, str] | None = None
    license: list[str] | None = None
    type: str | None = None
    notification_url: str | None = None
    bin: list[str] | None = None
    authors: list[PhpComposerAuthors] | None = None
    description: str | None = None
    homepage: str | None = None
    keywords: list[str] | None = None
    time: str | None = None


# PhpComposerInstalledEntry is an alias for PhpComposerLockEntry
PhpComposerInstalledEntry = PhpComposerLockEntry


class PhpPeclEntry(BaseModel):
    name: str
    version: str
    license: list[str] | None = None



from fluid_sbom.pkg.python import (
    PythonFileDigest,
    PythonFileRecord,
)
import os
from pathlib import (
    Path,
)
from typing import (
    TextIO,
)


def parse_wheel_or_egg_record(reader: TextIO) -> list[PythonFileRecord]:
    file_records = []
    lines = reader.read().strip().split("\n")
    for line in lines:
        parts = line.split(",")
        if len(parts) != 3:
            continue  # Skip invalid lines

        path = parts[0]
        digest_part = parts[1].split("=")
        if len(digest_part) != 2:
            continue  # Skip invalid digest format

        algorithm = digest_part[0]
        value = digest_part[1]
        size = int(parts[2])

        digest = PythonFileDigest(algorithm=algorithm, value=value)
        record = PythonFileRecord(path=path, digest=digest, size=str(size))
        file_records.append(record)

    return file_records


def parse_installed_files(
    reader: TextIO, location: str, site_packages_root_path: str
) -> list[PythonFileRecord]:
    installed_files = []
    for line in reader.readlines():
        line = line.rstrip("\n")
        if location and site_packages_root_path:
            joined_path = os.path.join(Path(location).parent, line)
            line = str(Path(site_packages_root_path, joined_path).resolve())
        installed_record = PythonFileRecord(path=line)
        installed_files.append(installed_record)

    return installed_files



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
from more_itertools import (
    flatten,
)
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)
from typing import (
    cast,
)

LOGGER = logging.getLogger(__name__)


def parse_dotnet_package_lock(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json: IndexedDict = cast(
        IndexedDict, parse_json_with_tree_sitter(reader.read_closer.read())
    )
    packages: list[Package] = []
    relationships: list[Relationship] = []

    dependencies = {}
    for package_name, package_value in flatten(
        x.items() for x in package_json.get("dependencies", {}).values()
    ):
        version = package_value.get("resolved", "")
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = package_value.position.start.line
        dependencies[package_name] = package_value.get("dependencies", {})
        try:
            packages.append(
                Package(
                    name=package_name,
                    version=version,
                    locations=[location],
                    licenses=[],
                    type=PackageType.DotnetPkg,
                    language=Language.DOTNET,
                    metadata=None,
                    p_url=PackageURL(
                        type="nuget",
                        namespace="",
                        name=package_name,
                        version=version,
                        qualifiers={},
                        subpath="",
                    ).to_string(),
                )
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types "
                "are incorrect.",
                extra={
                    "extra": {
                        "exception": ex.errors(include_url=False),
                        "location": location.path(),
                    }
                },
            )
            continue

    for package_name, depens in dependencies.items():
        deps = (
            p
            for p in packages
            for dep_name in depens.keys()
            if dep_name == p.name
        )
        if current_package := next(
            (x for x in packages if x.name == package_name), None
        ):
            relationships.extend(
                Relationship(
                    from_=x,
                    to_=current_package,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    data=None,
                )
                for x in deps
            )

    return packages, relationships



from __future__ import (
    annotations,
)

from fluid_sbom.pkg.cataloger.redhat.rpmdb.berkeley.constants import (
    NO_ENCRYPTION_ALGORITHM,
)
from pydantic import (
    BaseModel,
)
import struct


class GenericMetadataPage(BaseModel):
    lsn: bytes
    page_no: int
    magic: int
    version: int
    page_size: int
    encryption_alg: int
    page_type: int
    meta_flags: int
    unused_1: int
    free: int
    last_page_no: int
    n_parts: int
    key_count: int
    record_count: int
    flags: bytes
    unique_file_id: int | None = None

    @classmethod
    def from_bytes(  # pylint: disable=too-many-locals
        cls,
        data: bytes,
        byte_order: str,
    ) -> GenericMetadataPage:
        fmt = f"{'>' if byte_order == 'big' else '<'}8sIIIIBBBBIIIII20s"
        expected_size = struct.calcsize(fmt)
        if len(data) < expected_size:
            raise ValueError(
                "Data too short, expected at least "
                f"{expected_size} bytes, got {len(data)} bytes"
            )

        try:
            unpacked_data = struct.unpack(fmt, data[:expected_size])
        except struct.error as exc:
            raise ValueError(
                f"Failed to unpack GenericMetadataPage: {exc}"
            ) from exc

        # Unpack the fields
        (
            lsn,
            page_no,
            magic,
            version,
            page_size,
            encryption_alg,
            page_type,
            meta_flags,
            unused_1,
            free,
            last_page_no,
            n_parts,
            key_count,
            record_count,
            flags,
            # UniqueFileID,
        ) = unpacked_data

        return cls(
            lsn=lsn,
            page_no=page_no,
            magic=magic,
            version=version,
            page_size=page_size,
            encryption_alg=encryption_alg,
            page_type=page_type,
            meta_flags=meta_flags,
            unused_1=unused_1,
            free=free,
            last_page_no=last_page_no,
            n_parts=n_parts,
            key_count=key_count,
            record_count=record_count,
            flags=flags,
            # UniqueFileID,
        )

    @classmethod
    def validate(cls, value: GenericMetadataPage) -> None:  # type: ignore
        """
        Validates the parsed GenericMetadataPage.

        :raises ValueError: If the encryption algorithm is not supported.
        """
        if value.encryption_alg != NO_ENCRYPTION_ALGORITHM:
            raise ValueError(
                f"Unexpected encryption algorithm: {value.encryption_alg}"
            )

    def __repr__(self) -> str:
        return (
            "GenericMetadataPage("
            f"LSN={self.lsn.hex()}, PageNo={self.page_no}, Magic={self.magic},"
            f" Version={self.version}, PageSize={self.page_size}, "
            f"EncryptionAlg={self.encryption_alg}, PageType={self.page_type}, "
            f"MetaFlags={self.meta_flags}, Unused1={self.unique_file_id}, "
            f"Free={self.free}, LastPageNo={self.last_page_no}, "
            f"NParts={self.n_parts}, "
            f"KeyCount={self.key_count}, RecordCount={self.record_count}, "
            f"Flags={self.flags.hex()}, UniqueFileID={self.unique_file_id})"
        )


def parse_generic_metadata_page(data: bytes) -> GenericMetadataPage:
    return GenericMetadataPage.from_bytes(data, "little")



from fluid_sbom.pkg.cataloger.dotnet.parse_csproj import (
    parse_csproj,
)
from fluid_sbom.pkg.cataloger.dotnet.parse_dotnet_package_config import (
    parse_dotnet_pkgs_config,
)
from fluid_sbom.pkg.cataloger.dotnet.parse_dotnet_package_lock import (
    parse_dotnet_package_lock,
)
from fluid_sbom.pkg.cataloger.generic.cataloger import (
    Request,
)
from fnmatch import (
    fnmatch,
)
import reactivex
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)


def on_next_dotnet(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if any(
                    fnmatch(value, x)
                    for x in (
                        "**/packages.config",
                        "packages.config",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_dotnet_pkgs_config,
                            parser_name="dotnet-parse-packages-config",
                        )
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        "**/packages.lock.json",
                        "packages.lock.json",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_dotnet_package_lock,
                            parser_name="dotnet-parse-package-lock",
                        )
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        "**/project.csproj",
                        "project.csproj",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_csproj,
                            parser_name="dotnet-parse-csproj",
                        )
                    )
            except Exception as ex:  # pylint:disable=broad-exception-caught
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from fluid_sbom.pkg.java import (
    JavaPomProperties,
)
import re


def parse_pom_properties(
    path: str, file_content: str
) -> JavaPomProperties | None:
    prop_map = {}

    for line in file_content.splitlines():
        line = line.strip()
        # Skip empty lines and comments
        if line == "" or line.lstrip().startswith("#"):
            continue

        # Find the first occurrence of ':' or '='
        idx = next((i for i in range(len(line)) if line[i] in ":="), -1)
        if idx == -1:
            raise ValueError(
                f"Unable to split pom.properties key-value pairs: {line}"
            )

        key = line[:idx].strip()
        value = line[idx + 1 :].strip()
        prop_map[key] = value

    # Convert the dictionary to a JavaPomProperties object
    props = JavaPomProperties(path=path)
    for key, value in prop_map.items():
        key = re.sub(r"(?<!^)(?=[A-Z])", "_", key).lower()
        if hasattr(props, key):
            setattr(props, key, value)
        else:
            props.extra[  # pylint:disable=unsupported-assignment-operation
                key
            ] = value

    return props



from collections import (
    defaultdict,
)
from cyclonedx.factory.license import (
    LicenseFactory,
)
from cyclonedx.model import (
    Tool,
)
from cyclonedx.model.bom import (
    Bom,
)
from cyclonedx.model.component import (
    Component,
    ComponentType,
)
from cyclonedx.model.license import (
    LicenseExpression,
)
from cyclonedx.output import (
    make_outputter,
)
from cyclonedx.output.json import (
    JsonV1Dot5,
)
from cyclonedx.schema import (
    OutputFormat,
    SchemaVersion,
)
from cyclonedx.validation import (
    make_schemabased_validator,
)
from cyclonedx.validation.json import (
    JsonStrictValidator,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.config.config import (
    SbomConfig,
)
from fluid_sbom.format.common import (
    set_namespace_version,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.utils.exceptions import (
    CycloneDXValidationError,
)
import logging
from packageurl import (
    PackageURL,
)
from typing import (
    cast,
    TYPE_CHECKING,
)

LOGGER = logging.getLogger(__name__)


if TYPE_CHECKING:
    from cyclonedx.output.json import (
        Json as JsonOutputter,
    )
    from cyclonedx.output.xml import (
        Xml as XmlOutputter,
    )
    from cyclonedx.validation.xml import (
        XmlValidator,
    )


def format_cyclone_json(bom: Bom, output: str) -> None:
    file_path = f"{output}.json"
    json_output: "JsonOutputter" = JsonV1Dot5(bom)
    serialized_json = json_output.output_as_string()

    json_validator = JsonStrictValidator(SchemaVersion.V1_5)
    validation_error = json_validator.validate_str(serialized_json)

    if validation_error:
        raise CycloneDXValidationError(validation_error)

    LOGGER.info("🆗 CycloneDx JSON valid, output file at %s", file_path)
    json_output.output_to_file(file_path, True, indent=2)
    LOGGER.info("✅ Output file successfully generated")


def format_cyclone_xml(bom: Bom, output: str) -> None:
    file_path = f"{output}.xml"
    xml_outputter: "XmlOutputter" = make_outputter(
        bom, OutputFormat.XML, SchemaVersion.V1_5
    )
    serialized_xml = xml_outputter.output_as_string()

    xml_validator: "XmlValidator" = make_schemabased_validator(
        xml_outputter.output_format, xml_outputter.schema_version
    )
    validation_error = xml_validator.validate_str(serialized_xml)

    if validation_error:
        raise CycloneDXValidationError(validation_error)

    LOGGER.info("🆗 CycloneDx XML valid, output file at %s", file_path)
    xml_outputter.output_to_file(file_path, True, indent=2)
    LOGGER.info("✅ Output file successfully generated")


def pkg_to_component(package: Package) -> Component:
    lc_factory = LicenseFactory()
    licenses = []
    for lic in package.licenses:
        item = lc_factory.make_from_string(lic)
        if not isinstance(item, LicenseExpression):
            licenses.append(item)
    return Component(
        type=ComponentType.LIBRARY,
        name=package.name,
        version=package.version,
        licenses=licenses,
        bom_ref=f"{package.name}@{package.version}",
        purl=PackageURL.from_string(package.p_url),
    )


def format_cyclonedx_sbom(  # pylint:disable=too-many-locals
    packages: list[Package],
    relationships: list[Relationship],
    file_format: str,
    output: str,
    config: SbomConfig,
) -> None:
    namespace, version = set_namespace_version(config=config)
    bom = Bom()
    bom.metadata.component = root_component = Component(
        name=namespace,
        type=ComponentType.APPLICATION,
        licenses=[],
        bom_ref="",
        version=version,
    )
    bom.metadata.tools.add(Tool(vendor="Fluid Attacks", name="Fluid-Sbom"))

    component_cache = {pkg: pkg_to_component(pkg) for pkg in packages}

    components = component_cache.values()
    for component in components:
        bom.components.add(component)
        bom.register_dependency(root_component, [component])

    dependency_map: dict[Component, list[Component]] = defaultdict(list)

    for relationship in relationships:
        to_pkg = component_cache.get(
            cast(Package, relationship.to_),
            pkg_to_component(cast(Package, relationship.to_)),
        )
        from_pkg = component_cache.get(
            cast(Package, relationship.from_),
            pkg_to_component(cast(Package, relationship.from_)),
        )

        dependency_map[to_pkg].append(from_pkg)

    for ref, depends_on_list in dependency_map.items():
        bom.register_dependency(ref, depends_on_list)

    match file_format:
        case "cyclonedx-json":
            format_cyclone_json(bom, output)
        case "cyclonedx-xml":
            format_cyclone_xml(bom, output)



import logging
import os

LOGGER = logging.getLogger(__name__)

try:
    CI_COMMIT_REF_NAME = os.environ["CI_COMMIT_REF_NAME"]
    CI_COMMIT_SHA = os.environ["CI_COMMIT_SHA"]
    CI_COMMIT_SHORT_SHA = CI_COMMIT_SHA[0:8]
    BASE_DIR = os.path.dirname(os.path.abspath(__file__))
except KeyError as ex:
    LOGGER.exception("Environment variable %s doesn't exist", ex.args[0])
    raise



from fluid_sbom.pkg.cataloger.generic.cataloger import (
    Request,
)
from fluid_sbom.pkg.cataloger.php.parse_composer_json import (
    parse_composer_json,
)
from fluid_sbom.pkg.cataloger.php.parse_composer_lock import (
    parse_composer_lock,
)
from fluid_sbom.pkg.cataloger.php.parse_installed_json import (
    parse_installed_json,
)
from fluid_sbom.pkg.cataloger.php.parse_serialized import (
    parse_pecl_serialized,
)
from fnmatch import (
    fnmatch,
)
import reactivex
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)


def on_next_php(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if any(
                    fnmatch(value, pattern)
                    for pattern in ("**/composer.lock", "composer.lock")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_composer_lock,
                            parser_name="parse-php-composer-lock",
                        )
                    )
                elif any(
                    fnmatch(value, pattern)
                    for pattern in ("**/composer.json", "composer.json")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_composer_json,
                            parser_name="parse-php-composer-json",
                        )
                    )
                elif any(
                    fnmatch(value, pattern)
                    for pattern in ("**/installed.json", "installed.json")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_installed_json,
                            parser_name="parse-php-installed-json",
                        )
                    )
                elif fnmatch(value, "**/php/.registry/.channel.*/*.reg"):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_pecl_serialized,
                            parser_name="parse-php-pecl-serialized",
                        )
                    )
            except Exception as ex:  # pylint:disable=broad-exception-caught
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from contextlib import (
    suppress,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb import (
    berkeley,
    sqlite,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.entry import (
    header_import,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.errors import (
    InvalidDBFormat,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.package import (
    get_nevra,
    PackageInfo,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.rpmdb_interface import (
    RpmDBInterface,
)


class RpmDB:  # pylint:disable=too-few-public-methods
    def __init__(self, database: RpmDBInterface) -> None:
        self.database = database

    def list_packages(
        self,
    ) -> list[PackageInfo]:
        packages: list[PackageInfo] = []
        for entry in self.database.read():
            try:
                index_entries = header_import(entry)
            except ValueError as exc:
                raise ValueError("Failed to import header") from exc
            if index_entries:
                package = get_nevra(index_entries)
                packages.append(package)
        return packages


def open_db(file_path: str) -> RpmDB | None:
    with suppress(InvalidDBFormat):
        return RpmDB(sqlite.open_sqlite(file_path))
    with suppress(InvalidDBFormat):
        return RpmDB(berkeley.BerkeleyDB(file_path))
    return None



from fluid_sbom.utils.file import (
    Digest,
)
from pydantic import (
    BaseModel,
)


class RpmFileRecord(BaseModel):
    path: str
    mode: int
    size: int
    digest: Digest
    username: str
    group_name: str | None
    flags: str


class RpmDBEntry(BaseModel):
    id_: str
    name: str
    version: str
    epoch: int | None
    arch: str
    release: str
    source_rpm: str
    size: int
    vendor: str
    modularitylabel: str
    files: list[RpmFileRecord]



# pylint:disable=no-value-for-parameter
import click
from concurrent.futures import (
    ThreadPoolExecutor,
)
from fluid_sbom.advisories.database import (
    DATABASE,
)
from fluid_sbom.config.bugsnag import (
    initialize_bugsnag,
)
from fluid_sbom.config.config import (
    SbomConfig,
    SourceType,
)
from fluid_sbom.config.logger import (
    configure_logger,
)
from fluid_sbom.context.image import (
    get_context,
)
from fluid_sbom.format import (
    format_sbom,
)
from fluid_sbom.internal.file_resolver.container_image import (
    ContainerImage,
)
from fluid_sbom.internal.file_resolver.directory import (
    Directory,
)
from fluid_sbom.internal.operations.package_operation import (
    package_operations_factory,
)
from fluid_sbom.pkg.cataloger.complete import (
    complete_package,
)
from fluid_sbom.sources.docker import (
    get_docker_image,
)
import logging
import os
import textwrap

LOGGER = logging.getLogger(__name__)


def show_banner() -> None:
    logo = textwrap.dedent(
        """
         ───── ⌝
        |    ⌝|  Fluid Attacks
        |  ⌝  |  We hack your software.
         ─────
        """
    )
    click.secho(logo, fg="red")


@click.command()
@click.argument("source")
@click.option(
    "--from",
    "o_from",
    type=click.Choice(
        ["docker", "dir", "docker-daemon"], case_sensitive=False
    ),
    help=(
        "Source of the scan: 'docker' for scanning Docker images "
        "or 'dir' for scanning directories."
    ),
    required=True,
)
@click.option(
    "--format",
    "output_format",
    type=click.Choice(
        [
            "fluid-json",
            "cyclonedx-json",
            "spdx-json",
            "cyclonedx-xml",
            "spdx-xml",
        ],
        case_sensitive=False,
    ),
    default="fluid-json",
    help="Output format for the scanned data.",
)
@click.option(
    "--output",
    "-o",
    help="Output filename.",
    default="sbom",
)
@click.option(
    "--docker-user",
    help="Docker registry username.",
)
@click.option(
    "--docker-password",
    help="Docker registry password.",
)
def scan(  # pylint:disable=too-many-arguments
    source: str,
    o_from: str,
    output_format: str,
    output: str,
    docker_user: str | None = None,
    docker_password: str | None = None,
) -> None:
    configure_logger()
    initialize_bugsnag()
    show_banner()
    config = SbomConfig(
        source=source,
        source_type=SourceType.from_string(o_from),
        output_format=output_format,
        output=output,
        resolver=None,
    )

    match config.source_type:
        case SourceType.DIRECTORY:
            config.resolver = Directory(root=source)
        case SourceType.DOCKER | SourceType.DOCKER_DAEMON:
            daemon = config.source_type == SourceType.DOCKER_DAEMON
            docker_image = get_docker_image(
                source,
                username=docker_user,
                password=docker_password,
                daemon=daemon,
            )
            if not docker_image:
                raise ValueError(f"No image found for {source}")

            context = get_context(
                docker_image,
                username=docker_user,
                password=docker_password,
                daemon=daemon,
            )
            if context is None:
                raise ValueError(f"No context found for {docker_image}")
            config.resolver = ContainerImage(
                img=docker_image, context=context, lazy=False
            )
        case _:
            raise ValueError(f"Unknown source: {source}")

    DATABASE.initialize()
    LOGGER.info("📦 Generating SBOM from %s: %s", o_from, source)

    packages, relationships = package_operations_factory(config.resolver)
    with ThreadPoolExecutor(
        max_workers=min(
            32, (os.cpu_count() or 1) * 5 if os.cpu_count() is not None else 32
        )
    ) as executor:
        LOGGER.info("📦 Gathering additional package information")
        packages = list(filter(None, executor.map(complete_package, packages)))

    LOGGER.info("📦 Preparing %s report", config.output_format)
    format_sbom(packages, relationships, config)


if __name__ == "__main__":
    scan(prog_name="sbom")



from __future__ import (
    annotations,
)

from fluid_sbom.pkg.cataloger.redhat.rpmdb.berkeley.constants import (
    HASH_MAGIC_NUMBER,
    HASH_MAGIC_NUMBER_BE,
    HASH_METADATA_PACKAGE_TYPE,
)
from fluid_sbom.pkg.cataloger.redhat.rpmdb.berkeley.generic_page import (
    GenericMetadataPage,
)
from pydantic import (
    BaseModel,
)
import struct


class HashMetadata(BaseModel):
    """
    Represents the Hash Metadata in the RPM database.
    """

    generic_metadata_page: GenericMetadataPage
    max_bucket: int
    high_mask: int
    low_mask: int
    fill_factor: int
    num_keys: int
    char_key_hash: int

    @classmethod
    def from_bytes(cls, data: bytes, byte_order: str) -> "HashMetadata":
        expected_size = 96  # Up to CharKeyHash (bytes 0-95)
        if len(data) < expected_size:
            raise ValueError(
                f"Data too short, expected at least {expected_size} bytes,"
                " got {len(data)} bytes"
            )

        # Parse the GenericMetadataPage part
        generic_metadata = GenericMetadataPage.from_bytes(
            data[:72], byte_order
        )

        # Parse the additional fields (MaxBucket to CharKeyHash)
        fmt = f"{'>' if byte_order == 'big' else '<'}6I"
        additional_fields_size = struct.calcsize(fmt)
        try:
            unpacked_data = struct.unpack(
                fmt, data[72 : 72 + additional_fields_size]
            )
        except struct.error as exc:
            raise ValueError(
                f"Failed to unpack HashMetadata additional fields: {exc}"
            ) from exc

        (
            max_bucket,
            high_mask,
            low_mask,
            fill_factor,
            num_keys,
            char_key_hash,
        ) = unpacked_data

        return cls(
            generic_metadata_page=generic_metadata,
            max_bucket=max_bucket,
            high_mask=high_mask,
            low_mask=low_mask,
            fill_factor=fill_factor,
            num_keys=num_keys,
            char_key_hash=char_key_hash,
        )

    @classmethod
    def validate(cls, value: HashMetadata) -> None:  # type: ignore
        """
        Validates the HashMetadata.

        :raises ValueError: If validation fails.
        """
        GenericMetadataPage.validate(value.generic_metadata_page)

        if value.generic_metadata_page.magic != HASH_MAGIC_NUMBER:
            raise ValueError(
                "Unexpected DB magic number: "
                f"{hex(value.generic_metadata_page.magic)}"
            )

        if value.generic_metadata_page.page_type != HASH_METADATA_PACKAGE_TYPE:
            raise ValueError(
                "Unexpected page type: "
                + str(value.generic_metadata_page.page_type)
            )


class HashMetadataPage:
    """
    Represents a Hash Metadata Page with endian information.
    """

    def __init__(self, hash_metadata: HashMetadata, swapped: bool):
        self.hash_metadata = hash_metadata
        self.swapped = (
            swapped  # Indicates if the byte order is swapped (big-endian)
        )

    @classmethod
    def from_bytes(cls, data: bytes) -> "HashMetadataPage":
        """
        Parses a HashMetadataPage from a bytes object.

        :param data: The bytes object containing the page data.
        :return: An instance of HashMetadataPage.
        :raises ValueError: If the data is invalid or parsing fails.
        """
        swapped = False
        byte_order = "little"

        # First attempt to parse using little-endian byte order
        hash_metadata = HashMetadata.from_bytes(data, byte_order)

        if hash_metadata.generic_metadata_page.magic == HASH_MAGIC_NUMBER_BE:
            # Re-read the GenericMetadataPage using big-endian byte order
            swapped = True
            byte_order = "big"

            # Re-parse only the GenericMetadataPage portion
            generic_metadata = GenericMetadataPage.from_bytes(
                data[:72], byte_order
            )

            # Create a new HashMetadata instance with the corrected
            # GenericMetadataPage
            metadata = HashMetadata(
                generic_metadata_page=generic_metadata,
                max_bucket=hash_metadata.max_bucket,
                high_mask=hash_metadata.high_mask,
                low_mask=hash_metadata.low_mask,
                fill_factor=hash_metadata.fill_factor,
                num_keys=hash_metadata.num_keys,
                char_key_hash=hash_metadata.char_key_hash,
            )
            hash_metadata = metadata

        # Validate the metadata
        HashMetadata.validate(hash_metadata)

        return cls(hash_metadata, swapped)

    def __repr__(self) -> str:
        return (
            f"HashMetadataPage(hash_metadata={self.hash_metadata}, "
            f"swapped={self.swapped})"
        )


def parse_hash_metadata_page(data: bytes) -> HashMetadataPage:
    """
    Parses a HashMetadataPage from a bytes object.

    :param data: The bytes object containing the page data.
    :return: An instance of HashMetadataPage.
    :raises ValueError: If the data is invalid or parsing fails.
    """
    return HashMetadataPage.from_bytes(data)



from contextlib import (
    suppress,
)
from datetime import (
    datetime,
    timezone,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.alpm import (
    AlpmDBEntry,
    AlpmFileRecord,
)
from fluid_sbom.pkg.cataloger.arch.package import (
    new_package,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.utils.file import (
    Digest,
)
import gzip
import os
from pathlib import (
    Path,
)
from typing import (
    Any,
    TextIO,
)

IGNORED_FILES = {
    "/set": True,
    ".BUILDINFO": True,
    ".PKGINFO": True,
    "": True,
}


def parse_pkg_files(pkg_fields: dict[str, Any]) -> AlpmDBEntry | None:
    entry = AlpmDBEntry()

    entry.licenses = pkg_fields.get("license", "")
    entry.base_package = pkg_fields.get("base", "")
    entry.package = pkg_fields.get("name", "")
    entry.version = pkg_fields.get("version", "")
    entry.description = pkg_fields.get("desc", "")
    entry.architecture = pkg_fields.get("arch", "")
    entry.size = pkg_fields.get("size", 0)
    entry.packager = pkg_fields.get("packager", "")
    entry.url = pkg_fields.get("url", "")
    entry.validation = pkg_fields.get("validation", "")
    entry.reason = pkg_fields.get("reason", 0)
    entry.files = list(
        AlpmFileRecord(path=item["path"], digests=item.get("digests"))
        for item in pkg_fields.get("files", [])
    )
    entry.backup = list(
        AlpmFileRecord(path=item["path"], digests=item["digests"])
        for item in pkg_fields.get("backup", [])
    )

    if not entry.package and not entry.files and not entry.backup:
        return None
    return entry


def parse_key_value_pair(line: str) -> dict[str, Any]:
    try:
        key, value = line.split("\n", 1)
    except ValueError:
        return {}
    key = key.replace("%", "").lower()
    value = value.strip()

    if key == "files":
        return {key: _parse_files(value)}
    if key == "backup":
        return {key: parse_backup(value)}
    if key in ["reason", "size"]:
        return {key: parse_numeric_field(key, value)}
    return {key: value}


def _parse_files(value: str) -> list[dict[str, Any]]:
    files = []
    for item in value.split("\n"):
        path = f"/{item}"
        if path not in IGNORED_FILES:
            files.append({"path": path})
    return files


def parse_backup(value: str) -> list[dict[str, Any]]:
    backup = []
    for item in value.split("\n"):
        backup_fields = item.split("\t", 1)
        path = f"/{backup_fields[0]}"
        if path not in IGNORED_FILES:
            backup.append(
                {
                    "path": path,
                    "digests": [
                        Digest(algorithm="md5", value=backup_fields[1])
                    ],
                }
            )
    return backup


def parse_numeric_field(key: str, value: str) -> int:
    try:
        return int(value)
    except ValueError as exc:
        raise ValueError(f"Failed to parse {key} to integer: {value}") from exc


def parse_alpm_db_entry(reader: TextIO) -> AlpmDBEntry | None:
    pkg_fields: dict[str, Any] = {}
    lines = reader.read().split("\n\n")
    for line in lines:
        if not line.strip():
            break  # End of block or file
        pkg_fields.update(parse_key_value_pair(line))

    return parse_pkg_files(pkg_fields)


def parse_mtree(reader: TextIO) -> list[AlpmFileRecord]:
    file_info: dict[str, dict[str, str]] = {}
    result_records: list[AlpmFileRecord] = []
    data = reader.read()

    for line in data.splitlines():
        if line.startswith(".") and "time=" in line:
            parts = line.strip().split()
            file_path = parts[0][
                1:
            ]  # Remove the leading '.' from the file path
            file_attributes = {}
            for part in parts[
                1:
            ]:  # Skip the first part, which is the file path
                if "=" in part:
                    key, value = part.split("=", 1)
                    file_attributes[key] = value
            file_info[file_path] = file_attributes
    for file_path, file_attributes in file_info.items():
        if file_path.startswith("/."):
            continue
        result = AlpmFileRecord(
            path=file_path,
            size=file_attributes.get("size"),
            time=datetime.fromtimestamp(
                float(file_attributes["time"]), tz=timezone.utc
            ),
            type=file_attributes.get("type"),
            digests=[
                Digest(
                    algorithm=key.replace("digest", ""),
                    value=value,
                )
                for key, value in file_attributes.items()
                if key.endswith("digest")
            ],
            uid=file_attributes.get("uid"),
            gid=file_attributes.get("gid"),
            link=file_attributes.get("link"),
        )
        result_records.append(result)
    return result_records


def parse_alpm_db(
    _resolver: Resolver, _env: Environment, reader: LocationReadCloser
) -> tuple[list[Package], list[Relationship]]:
    data = parse_alpm_db_entry(reader.read_closer)
    if not data or not reader.location.coordinates:
        return ([], [])
    with suppress(IndexError):
        mtree_location = _resolver.files_by_path(
            os.path.join(
                Path(reader.location.coordinates.real_path).parent, "mtree"
            )
        )[0]
        if mtree_reader := _resolver.file_contents_by_location(
            mtree_location, function_reader=gzip.open, mode="rt"
        ):
            data.files = parse_mtree(mtree_reader)
    with suppress(IndexError):
        files_location = _resolver.files_by_path(
            os.path.join(
                Path(reader.location.coordinates.real_path).parent, "files"
            )
        )[0]
        if (
            files_reader := _resolver.file_contents_by_location(files_location)
        ) and (files_metadata := parse_alpm_db_entry(files_reader)):
            data.backup = files_metadata.backup

    package = new_package(data, _env.linux_release, reader.location)

    return [package] if package else [], []



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.ruby.package import (
    package_url,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
import os
from pydantic import (
    ValidationError,
)
from tree_sitter import (
    Language as TLanguage,
    Node,
    Parser,
)

LOGGER = logging.getLogger(__name__)


def collect_gem_entries(_content: str) -> list[Node]:
    parser = Parser()
    parser.set_language(
        TLanguage(
            os.path.join(
                os.environ["TREE_SITTETR_PARSERS_DIR"],
                "gemfilelock.so",
            ),
            "gemfilelock",
        )
    )
    result = parser.parse(_content.encode("utf-8"))
    gem_section = next(
        (
            node
            for node in result.root_node.children
            if node.type == "gem_section"
        ),
        None,
    )
    if gem_section and (
        specs := next(
            (x for x in gem_section.children[1].children if x.type == "specs"),
            None,
        )
    ):
        return [x.children[0] for x in specs.children[1:]]

    return []


def parse_gemfile_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []
    gem_entries = collect_gem_entries(reader.read_closer.read())
    for gem_entry in gem_entries:
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = gem_entry.start_point[0] + 1
        gem_name = gem_entry.named_children[0].text.decode("utf-8")
        gem_version = gem_entry.named_children[1].text.decode("utf-8")
        try:
            packages.append(
                Package(
                    name=gem_name,
                    version=gem_version,
                    locations=[location],
                    language=Language.RUBY,
                    licenses=[],
                    p_url=package_url(gem_name, gem_version),
                    type=PackageType.GemPkg,
                    metadata=None,
                )
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types "
                " are incorrect.",
                extra={
                    "extra": {
                        "exception": ex.errors(include_url=False),
                        "location": location.path(),
                    }
                },
            )
            continue
    for gem_entry in gem_entries:
        gem_entry_name = gem_entry.named_children[0].text.decode("utf-8")
        _package = next(
            (x for x in packages if x.name == gem_entry_name), None
        )
        if not _package or not gem_entry.parent:
            continue
        for dependency_node in (
            x for x in gem_entry.parent.children if x.type == "dependency"
        ):
            dependency_name = dependency_node.named_children[0].text.decode(
                "utf-8"
            )
            if dependency_package := next(
                (x for x in packages if x.name == dependency_name), None
            ):
                relationships.append(
                    Relationship(
                        from_=_package,
                        to_=dependency_package,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        data=None,
                    )
                )
    return packages, relationships



from fluid_sbom.internal.cache import (
    dual_cache,
)
import requests
from typing import (
    Any,
)


@dual_cache
def get_pypi_package(
    package_name: str, version: str | None = None
) -> dict[str, Any] | None:
    url = (
        f"https://pypi.org/pypi/{package_name}"
        f"{'/' +version if version else ''}/json"
    )
    response = requests.get(url, timeout=20)
    if response.status_code == 200:
        package_info = response.json()
        return package_info

    return None



from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.toml import (
    parse_toml_with_tree_sitter,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.rust.package import (
    package_url,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.utils.exceptions import (
    UnexpectedNode,
)
import logging
from typing import (
    ItemsView,
)

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def _get_version(value: IndexedDict | str) -> str:
    if isinstance(value, str):
        return value
    if "git" in value:
        repo_url: str = value["git"]
        branch: str = value["branch"]
        return f"{repo_url}@{branch}"
    version: str = value["version"]
    return version


def _get_packages(
    reader: LocationReadCloser,
    dependencies: IndexedDict | None,
) -> list[Package]:
    if dependencies is None:
        return []

    general_location = _get_location(
        reader.location, dependencies.position.start.line
    )
    items: ItemsView[str, IndexedDict | str] = dependencies.items()
    return [
        Package(
            name=name,
            version=_get_version(value),
            locations=[
                (
                    _get_location(reader.location, value.position.start.line)
                    if isinstance(value, IndexedDict)
                    else general_location
                )
            ],
            language=Language.RUST,
            licenses=[],
            p_url=package_url(name, _get_version(value)),
            type=PackageType.RustPkg,
        )
        for name, value in items
    ]


def parse_cargo_toml(
    _: Resolver | None, __: Environment | None, reader: LocationReadCloser
) -> tuple[list[Package], list[Relationship]]:
    try:
        content = parse_toml_with_tree_sitter(reader.read_closer.read())
    except UnexpectedNode:
        return [], []

    deps: IndexedDict | None = content.get("dependencies")
    dev_deps: IndexedDict | None = content.get("dev-dependencies")
    packages = [
        *_get_packages(reader, deps),
        *_get_packages(reader, dev_deps),
    ]
    return packages, []



from fluid_sbom.utils.file import (
    Digest,
)
from pydantic import (
    BaseModel,
)


class ApkFileRecord(BaseModel):
    path: str
    owner_uid: str | None = None
    owner_gid: str | None = None
    permissions: str | None = None
    digest: Digest | None = None


class ApkDBEntry(BaseModel):
    package: str
    origin_package: str | None
    maintainer: str | None
    version: str
    architecture: str | None
    url: str | None
    description: str | None
    size: str
    installed_size: str | None
    dependencies: list[str]
    provides: list[str]
    checksum: str | None
    git_commit: str | None
    files: list[ApkFileRecord]



from fluid_sbom.linux.release import (
    Release,
)
from packageurl import (
    PackageURL,
)


def package_url(
    *,
    name: str,
    arch: str | None,
    epoch: int | None,
    source_rpm: str,
    version: str,
    release: str,
    distro: Release | None,
) -> str:
    namespace = ""
    if distro:
        namespace = distro.id_
    qualifiers: dict[str, str] = {}
    if arch:
        qualifiers["arch"] = arch
    if epoch:
        qualifiers["epoch"] = str(epoch)
    if source_rpm:
        qualifiers["upstream"] = source_rpm

    return PackageURL(
        type="rpm",
        namespace=namespace,
        name=name,
        version=f"{version}-{release}",
        qualifiers=qualifiers,
        subpath="",
    ).to_string()



from fluid_sbom.config.config import (
    SbomConfig,
    SourceType,
)
from fluid_sbom.internal.file_resolver.container_image import (
    ContainerImage,
)
from fluid_sbom.internal.file_resolver.directory import (
    Directory,
)
from fluid_sbom.pkg.package import (
    Package,
)


def merge_packages(packages: list[Package]) -> list[Package]:
    merged_packages: dict = {}

    for package in packages:
        if package.id_:
            if package.id_ in merged_packages:
                merged_packages[package.id_].locations.extend(
                    x
                    for x in package.locations
                    if x not in merged_packages[package.id_].locations
                )
            else:
                merged_packages[package.id_] = package
        else:
            merged_packages[package] = package

    return list(merged_packages.values())


def process_packages(packages: list[Package]) -> list[Package]:
    packages = merge_packages(packages)
    return packages


def set_namespace_version(config: SbomConfig) -> tuple[str, str | None]:
    namespace = ""
    version = None
    if config.source_type == SourceType.DIRECTORY and isinstance(
        config.resolver, Directory
    ):
        namespace = config.resolver.root
    if config.source_type == SourceType.DOCKER and isinstance(
        config.resolver, ContainerImage
    ):
        namespace = config.resolver.context.image_ref
        version = config.resolver.context.id
    return namespace, version



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.php.package import (
    new_package_from_composer,
)
from fluid_sbom.pkg.package import (
    Package,
)
import logging
from typing import (
    cast,
)

LOGGER = logging.getLogger(__name__)


def parse_composer_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    try:
        package_json: IndexedDict = cast(
            IndexedDict, parse_json_with_tree_sitter(reader.read_closer.read())
        )
    except ValueError:
        LOGGER.warning(
            "Failed to parse package.json",
            extra={
                "extra": {
                    "location": reader.location,
                }
            },
        )
        return [], []
    packages: list[Package] = []
    relationships: list[Relationship] = []

    for is_dev, package in [
        *[(False, x) for x in package_json.get("packages", [])],
        *[(True, x) for x in package_json.get("packages-dev", [])],
    ]:
        new_location = deepcopy(reader.location)

        if pkg := new_package_from_composer(package, new_location, is_dev):
            packages.append(pkg)

    for package in packages:
        for dep_name in package.metadata.require or []:
            package_dep = next(
                (x for x in packages if x.name == dep_name), None
            )
            if package_dep:
                relationships.append(
                    Relationship(
                        from_=package,
                        to_=package_dep,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        data=None,
                    )
                )
    return packages, relationships



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
    Position,
)
from fluid_sbom.internal.collection.yaml import (
    parse_yaml_with_tree_sitter,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.javascript.package import (
    package_url,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.npm import (
    PnpmEntry,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.utils.file import (
    Digest,
)
import logging
from pydantic import (
    ValidationError,
)
import re
from typing import (
    cast,
)

LOGGER = logging.getLogger(__name__)


def extract_package_name_from_key_dependency(item: str) -> str | None:
    # Regex pattern to extract the package name
    pattern = r"^@?[\w-]+/[\w-]+$"
    match = re.match(pattern, item)
    if match:
        return match.group(0)
    return None


def extract_version_from_value_dependency(item: str) -> str | None:
    # Regex pattern to extract the version number before any parentheses
    pattern = r"^(\d+\.\d+\.\d+)"
    match = re.match(pattern, item)
    if match:
        return match.group(1)
    return None


def _get_package(
    packages: list[Package], dep_name: str | None, dep_version: str | None
) -> Package | None:
    return next(
        (
            x
            for x in packages
            if x.name == dep_name and x.version == dep_version
        ),
        None,
    )


def _generate_relations_relationship(
    package_yaml: IndexedDict, packages: list[Package]
) -> list[Relationship]:
    relationships: list[Relationship] = []
    for package_key, package_value in package_yaml["packages"].items():
        if match_ := re.search(r"/(@?[^@]+)@(\d+\.\d+\.\d+)", package_key):
            package_name = match_.groups()[0]
            package_version = match_.groups()[1]
            current_package = _get_package(
                packages, dep_name=package_name, dep_version=package_version
            )
            if dependencies := package_value.get("dependencies"):
                for dep_name, dep_version in dependencies.items():
                    dep_name = extract_package_name_from_key_dependency(
                        dep_name
                    )
                    dep_version = extract_version_from_value_dependency(
                        dep_version
                    )
                    if dep := _get_package(
                        packages,
                        dep_name,
                        dep_version,
                    ):
                        relationships.append(
                            Relationship(
                                from_=dep,
                                to_=current_package,
                                type=(
                                    RelationshipType.DEPENDENCY_OF_RELATIONSHIP
                                ),
                                data=None,
                            )
                        )
    return relationships


def parse_pnpm_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_yaml: IndexedDict = cast(
        IndexedDict, parse_yaml_with_tree_sitter(reader.read_closer.read())
    )
    if not package_yaml:
        LOGGER.warning(
            "Failed to parse pnpm lock file",
            extra={
                "extra": {
                    "location": reader.location.path(),
                },
            },
        )
        return [], []
    packages: list[Package] = []
    relationships: list[Relationship] = []
    for package_key, _package_value in package_yaml["packages"].items():
        if match_ := re.search(r"/(@?[^@]+)@(\d+\.\d+\.\d+)", package_key):
            package_name = match_.groups()[0]
            package_version = match_.groups()[1]
            current_location: Location = deepcopy(reader.location)
            if current_location.coordinates:
                position: Position = package_yaml["packages"].get_key_position(
                    package_key
                )
                current_location.coordinates.line = position.start.line
            try:
                packages.append(
                    Package(
                        name=package_name,
                        version=package_version,
                        locations=[current_location],
                        language=Language.JAVASCRIPT,
                        licenses=[],
                        type=PackageType.NpmPkg,
                        p_url=package_url(package_name, package_version),
                        metadata=PnpmEntry(
                            is_dev=_package_value.get("dev", False),
                            integrity=Digest(
                                algorithm="sha-512",
                                value=_package_value["resolution"][
                                    "integrity"
                                ],
                            )
                            if "resolution" in _package_value
                            and "integrity" in _package_value["resolution"]
                            else None,
                        ),
                    )
                )
            except ValidationError as ex:
                LOGGER.warning(
                    "Malformed package. Required fields are missing or data "
                    "types are incorrect.",
                    extra={
                        "extra": {
                            "exception": ex.errors(include_url=False),
                            "location": current_location.path(),
                        }
                    },
                )
                continue
    relationships = _generate_relations_relationship(package_yaml, packages)
    return packages, relationships



from bs4 import (
    BeautifulSoup,
)
from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


def parse_csproj(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    root = BeautifulSoup(reader.read_closer.read(), features="html.parser")
    for pkg in root.find_all("packagereference", recursive=True):
        line = pkg.sourceline
        if (id_ := pkg.get("include")) and (version := pkg.get("version")):
            location = deepcopy(reader.location)
            if location.coordinates:
                location.coordinates.line = line
        try:
            packages.append(
                Package(
                    name=id_,
                    version=version,
                    locations=[location],
                    language=Language.DOTNET,
                    licenses=[],
                    type=PackageType.DotnetPkg,
                    metadata=None,
                    p_url=PackageURL(
                        type="nuget",
                        namespace="",
                        name=id_,
                        version=version,
                        qualifiers={},
                        subpath="",
                    ).to_string(),
                )
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types "
                "are incorrect.",
                extra={
                    "extra": {
                        "exception": ex.errors(include_url=False),
                        "location": location.path(),
                    }
                },
            )
            continue

    return packages, []



from bs4 import (
    BeautifulSoup,
)
from contextlib import (
    suppress,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.javascript.package import (
    package_url,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import re

SCRIPT_DEP = re.compile(
    r"(?P<name>[^\s\/]*)(?P<separator>[-@\/])"
    r"(?P<version>(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*))"
)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        return location.model_copy(
            update={
                "coordinates": location.coordinates.model_copy(
                    update={"line": sourceline}
                )
            }
        )
    return location


def parse_html_scripts(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    html = None
    with suppress(UnicodeError):
        try:
            html = BeautifulSoup(reader.read_closer, features="html.parser")
        except AssertionError:
            return [], []

    if not html:
        return [], []

    packages = [
        Package(
            name=matched.group("name"),
            version=matched.group("version"),
            licenses=[],
            locations=[_get_location(reader.location, script.sourceline)],
            language=Language.JAVASCRIPT,
            type=PackageType.NpmPkg,
            p_url=package_url(matched.group("name"), matched.group("version")),
        )
        for script in html("script")
        if (
            (src_attribute := script.attrs.get("src"))
            and src_attribute.endswith(".js")
            and (matched := SCRIPT_DEP.search(src_attribute))
        )
    ]
    return packages, []



from pydantic import (
    BaseModel,
)


class Coordinates(BaseModel):
    real_path: str
    file_system_id: str | None = None
    line: int | None = None

    def __str__(self) -> str:
        result = f"RealPath={self.real_path}"
        if self.file_system_id:
            result += f" Layer={self.file_system_id}"
        return f"Location<{result}>"



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.elixir.package import (
    new_package,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.elixir import (
    ElixirMixLockEntry,
)
from fluid_sbom.pkg.package import (
    Package,
)
import os
from tree_sitter import (
    Language as TLanguage,
    Node,
    Parser,
)


def process_entry(entry: Node, location: Location) -> Package | None:
    name = next(
        x for x in entry.named_children if x.type == "package_name"
    ).text.decode("utf-8")[1:-1]
    version = next(
        x for x in entry.named_children if x.type == "version"
    ).text.decode("utf-8")[1:-1]
    pkg_hash = next(
        x for x in entry.named_children if x.type == "checksum"
    ).text.decode("utf-8")[1:-1]
    pkg_hash_ext = next(
        x for x in entry.named_children if x.type == "optional_checksum"
    ).text.decode("utf-8")[1:-1]
    new_location = deepcopy(location)
    if new_location.coordinates:
        new_location.coordinates.line = entry.start_point[0] + 1

    package = new_package(
        ElixirMixLockEntry(
            name=name,
            version=version,
            pkg_hash=pkg_hash,
            pkg_hash_ext=pkg_hash_ext,
        ),
        new_location,
    )

    return package if package else None


def collect_dependencies(
    packages: list[Package], package_entries: list[Node]
) -> list[Relationship]:
    relationships: list[Relationship] = []
    for entry in package_entries:
        current_package_name = next(
            x for x in entry.named_children if x.type == "package_name"
        ).text.decode("utf-8")[1:-1]
        for dependency_list in next(
            node
            for node in entry.named_children
            if node.type == "dependencies"
        ).named_children:
            dependencies = [
                next(
                    y for y in x.named_children if y.type == "atom"
                ).text.decode("utf-8")[1:]
                for x in dependency_list.named_children
                if x.type == "dependency"
            ]
            relationships.extend(
                Relationship(
                    from_=next(x for x in packages if x.name == dep_name),
                    to_=next(
                        x for x in packages if x.name == current_package_name
                    ),
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    data=None,
                )
                for dep_name in dependencies
                if next((x for x in packages if x.name == dep_name), None)
            )
    return relationships


def collect_entries(
    _content: str, location: Location
) -> tuple[list[Package], list[Relationship]]:
    parser = Parser()
    parser.set_language(
        TLanguage(
            os.path.join(
                os.environ["TREE_SITTETR_PARSERS_DIR"],
                "mix_lock.so",
            ),
            "mix_lock",
        )
    )
    result = parser.parse(_content.encode("utf-8"))
    package_entries = list(
        node
        for node in result.root_node.children
        if node.type == "package_entry"
    )

    packages = [
        package
        for x in package_entries
        if (package := process_entry(x, location)) is not None
    ]
    relations = collect_dependencies(packages, package_entries)

    return packages, relations


def parse_mix_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []

    parser = Parser()
    parser.set_language(
        TLanguage(
            os.path.join(
                os.environ["TREE_SITTETR_PARSERS_DIR"],
                "mix_lock.so",
            ),
            "mix_lock",
        )
    )
    result = parser.parse(reader.read_closer.read().encode("utf-8"))
    package_entries = list(
        node
        for node in result.root_node.children
        if node.type == "package_entry"
    )

    packages = [
        package
        for x in package_entries
        if (package := process_entry(x, reader.location))
    ]
    relationships = collect_dependencies(packages, package_entries)

    return packages, relationships



from contextlib import (
    suppress,
)
from copy import (
    deepcopy,
)
from fluid_sbom import (
    advisories,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.internal.package_information.javascript import (
    get_npm_package,
    NPMPackage,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.npm import (
    NpmPackageLockEntry,
)
from fluid_sbom.pkg.package import (
    Artifact,
    HealthMetadata,
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.utils.file import (
    Digest,
)
from fluid_sbom.utils.package import (
    handle_licenses,
)
import logging
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


def new_package_lock_v1(
    location: Location,
    name: str,
    value: IndexedDict,
) -> Package | None:
    version: str = value.get("version", "")
    alias_prefix_package_lock = "npm:"
    if version.startswith(alias_prefix_package_lock):
        name, version = version.removeprefix(alias_prefix_package_lock).split(
            "@"
        )
    current_location = deepcopy(location)
    if current_location.coordinates:
        current_location.coordinates.line = value.position.start.line
    try:
        return Package(
            name=name,
            version=version,
            locations=[current_location],
            language=Language.JAVASCRIPT,
            licenses=[],
            type=PackageType.NpmPkg,
            metadata=NpmPackageLockEntry(
                resolved=value["resolved"],
                integrity=value["integrity"],
                is_dev=value.get("dev", False),
            )
            if value.get("resolved") and "integrity" in value
            else None,
            p_url=package_url(name, version),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types "
            "are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": location.path(),
                }
            },
        )
        return None


def new_package_lock_v2(
    location: Location,
    name: str,
    value: IndexedDict,
) -> Package | None:
    version: str = value.get("version", "")
    current_location = location

    if not name or not version:
        return None

    if current_location.coordinates:
        current_location.coordinates.line = value.position.start.line

    try:
        return Package(
            name=name,
            version=version,
            locations=[current_location],
            language=Language.JAVASCRIPT,
            licenses=[],
            type=PackageType.NpmPkg,
            metadata=NpmPackageLockEntry(
                resolved=value.get("resolved"),
                integrity=value.get("integrity"),
                is_dev=value.get("dev", False),
            ),
            p_url=package_url(name, version),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types "
            "are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": location.path(),
                }
            },
        )
        return None


def package_url(name: str, version: str) -> str:
    namespace = ""
    fields = name.split("/", 2)
    if len(fields) > 1:
        namespace = fields[0]
        name = fields[1]
    return PackageURL(
        type="npm",
        namespace=namespace,
        name=name,
        version=version,
        qualifiers={},
        subpath="",
    ).to_string()


def _set_author(npm_package: NPMPackage) -> str | None:
    author: str | None = None
    if "author" in npm_package:
        package_author = npm_package["author"]
        if isinstance(package_author, dict) and "name" in package_author:
            author = package_author["name"]
            if "email" in package_author:
                author = f'{author} <{package_author["email"]}>'
        elif package_author and isinstance(package_author, str):
            author = str(package_author)
        return author
    return None


def complete_package(package: Package) -> Package:
    pkg_advisories = advisories.get_package_advisories(package)
    if pkg_advisories:
        package.advisories = pkg_advisories

    npm_package = get_npm_package(package.name)
    if not npm_package:
        return package

    latest_version = None
    latest_version_created_at = None
    using_pre_release = "-" in package.version
    if using_pre_release:
        latest_version = list(npm_package["versions"].keys())[-1]
        latest_version_created_at = npm_package["time"]["modified"]
    elif npm_package.get("dist-tags"):
        latest_version = npm_package["dist-tags"]["latest"]
        latest_version_created_at = npm_package["time"][latest_version]

    current_package = npm_package["versions"].get(package.version)
    artifact = None
    if current_package:
        with suppress(KeyError):
            digest_value = (
                current_package.get("dist", {}).get("integrity") or None
            )
            artifact = Artifact(
                url=current_package["dist"]["tarball"],
                integrity=Digest(
                    algorithm="sha512" if digest_value else None,
                    value=digest_value,
                ),
            )

    package.health_metadata = HealthMetadata(
        latest_version=latest_version,
        latest_version_created_at=latest_version_created_at,
        artifact=artifact,
        authors=_set_author(npm_package),
    )

    licenses = npm_package.get("license") or None
    if licenses and isinstance(licenses, str | list | dict):
        package.licenses = handle_licenses(licenses)

    return package



from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.json import (
    parse_json_with_tree_sitter,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.swift.package import (
    new_swift_package_manager_package,
)
from fluid_sbom.pkg.package import (
    Package,
)
import logging
from typing import (
    cast,
)

LOGGER = logging.getLogger(__name__)


def parse_package_resolved(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    try:
        package_resolved: IndexedDict = cast(
            IndexedDict, parse_json_with_tree_sitter(reader.read_closer.read())
        )
    except ValueError:
        LOGGER.warning(
            "Failed to parse package.json",
            extra={
                "extra": {
                    "location": reader.location,
                }
            },
        )
        return [], []
    packages: list[Package] = []
    relationships: list[Relationship] = []

    for pin in package_resolved["pins"]:
        name = pin["identity"]
        version = pin["state"]["version"]
        new_location = deepcopy(reader.location)
        if new_location.coordinates:
            new_location.coordinates.line = pin.position.start.line

        if pkg := new_swift_package_manager_package(
            name,
            version,
            pin["location"],
            pin["state"]["revision"],
            new_location,
        ):
            packages.append(pkg)

    return packages, relationships



from bs4 import (
    BeautifulSoup,
    NavigableString,
    Tag,
)
from copy import (
    deepcopy,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.java.maven_repo_utils import (
    recursively_find_versions_from_parent_pom,
)
from fluid_sbom.pkg.cataloger.java.package import (
    package_url,
)
from fluid_sbom.pkg.java import (
    JavaArchive,
    JavaPomParent,
    JavaPomProject,
    JavaPomProperties,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
from pydantic import (
    BaseModel,
    ValidationError,
)
import re

LOGGER = logging.getLogger(__name__)


class ParsedPomProject(BaseModel):
    java_pom_project: JavaPomProject
    licenses: list[str]


def extract_bracketed_text(item: str) -> str:
    match = re.search(r"\$\{([^}]+)\}", item)
    if match:
        return match.group(1)
    return ""


def _get_text(parent: Tag | NavigableString, name: str) -> str | None:
    element = parent.find_next(name)
    if element:
        return element.get_text()
    return None


def new_package_from_pom_xml(
    project: Tag, dependency: Tag, location: Location
) -> Package | None:
    java_archive = JavaArchive(
        pom_properties=JavaPomProperties(
            group_id=_get_text(dependency, "groupid"),
            artifact_id=_get_text(dependency, "artifactid"),
            version=_get_text(dependency, "version")
            if dependency.version
            else None,
        )
    )
    name = _get_text(dependency, "artifactid")
    version = dependency.version.get_text() if dependency.version else None
    if (
        version
        and version.startswith("${")
        and (
            parent_version_node := project.find_next(
                extract_bracketed_text(version)
            )
        )
    ):
        version_text = parent_version_node.get_text()
        if version_text and not version_text.startswith("${"):
            version = version_text
    if (
        not version  # pylint:disable=too-many-boolean-expressions
        and java_archive.pom_properties
        and java_archive.pom_properties.group_id
        and java_archive.pom_properties.artifact_id
        and (parent := project.find_next("parent"))
        and (parent_groupid_node := parent.find_next("groupid"))
        and (parent_artifactid_node := parent.find_next("artifactid"))
        and (parent_version_node := parent.find_next("version"))
    ):
        version = recursively_find_versions_from_parent_pom(
            group_id=java_archive.pom_properties.group_id,
            artifact_id=java_archive.pom_properties.artifact_id,
            parent_group_id=parent_groupid_node.get_text(),
            parent_artifact_id=parent_artifactid_node.get_text(),
            parent_version=parent_version_node.get_text(),
        )

    if location.coordinates:
        location.coordinates.line = dependency.sourceline
    try:
        return Package(
            name=name or "",
            version=version or "",
            licenses=[],
            locations=[location],
            language=Language.JAVA,
            type=PackageType.JavaPkg,
            metadata=java_archive,
            p_url=package_url(name or "", version or "", java_archive),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types "
            "are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": location.path(),
                }
            },
        )
        return None


def parse_pom_xml(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    root = BeautifulSoup(reader.read_closer, features="html.parser")
    pkgs = []
    if (
        (project := root.project)
        and str(project.get("xmlns")) == "http://maven.apache.org/POM/4.0.0"
        and (dependencies := project.find("dependencies", recursive=False))
        and isinstance(dependencies, Tag)
    ):
        for dependency in dependencies.find_all("dependency"):
            pkg = new_package_from_pom_xml(
                project, dependency, deepcopy(reader.location)
            )
            if pkg:
                pkgs.append(pkg)
    return pkgs, []


def decode_pom_xml(content: str) -> Tag:
    return BeautifulSoup(content, features="html.parser")


def pom_parent(parent: Tag | None) -> JavaPomParent | None:
    if not parent:
        return None

    group_id = _get_text(parent, "groupId")
    artifact_id = _get_text(parent, "artifactId")
    version = _get_text(parent, "version")
    if not group_id or not artifact_id or not version:
        return None

    result = JavaPomParent(
        group_id=group_id,
        artifact_id=artifact_id,
        version=version,
    )

    if not result.group_id and not result.artifact_id and not result.version:
        return None

    return result


def parse_pom_xml_project(
    path: str, reader: str, _location: Location
) -> ParsedPomProject | None:
    project = BeautifulSoup(reader, features="xml").project
    if not project:
        return None
    return new_pom_project(path, project, _location)


def _find_direct_child(parent: Tag, tag: str) -> Tag | None:
    return next(
        (
            child
            for child in parent.find_all(tag, recursive=False)
            if child.parent == parent
        ),
        None,
    )


def new_pom_project(
    path: str, project: Tag, _location: Location
) -> ParsedPomProject:
    artifact_id = _safe_string(_find_direct_child(project, "artifactId"))
    name = _safe_string(_find_direct_child(project, "name"))
    project_url = _safe_string(_find_direct_child(project, "url"))

    licenses: list[str] = []
    if project.licenses:
        for license_ in project.licenses.find_all("license"):
            license_name: str | None = None
            license_url: str | None = None
            if name_node := license_.find_next("name"):
                license_name = name_node.get_text()
            elif url_node := license_.find_next("url"):
                license_url = url_node.get_text()

            if not license_name and not license_url:
                continue
            if license_name:
                licenses.append(license_name)
            elif license_url:
                licenses.append(license_url)

    return ParsedPomProject(
        java_pom_project=JavaPomProject(
            path=path,
            parent=pom_parent(_find_direct_child(project, "parent")),
            group_id=_safe_string(_find_direct_child(project, "groupId")),
            artifact_id=artifact_id,
            version=_safe_string(_find_direct_child(project, "version")),
            name=name,
            description=_safe_string(
                _find_direct_child(project, "description")
            ),
            url=project_url,
        ),
        licenses=licenses,
    )


def _safe_string(value: Tag | None) -> str:
    if not value:
        return ""
    return value.get_text()



from bugsnag.handlers import (
    BugsnagHandler,
)
import logging

LOGGER = logging.getLogger()


def configure_logger() -> None:
    logging.basicConfig(
        level=logging.INFO,
        format="[%(levelname)s] %(message)s",
    )
    handler = BugsnagHandler(extra_fields={"extra": ["extra"]})
    handler.setLevel(logging.WARNING)
    LOGGER.addFilter(handler.leave_breadcrumbs)
    LOGGER.addHandler(handler)



from datetime import (
    datetime,
    UTC,
)
from enum import (
    Enum,
)
from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.config.config import (
    SbomConfig,
)
from fluid_sbom.format.common import (
    set_namespace_version,
)
from fluid_sbom.format.types import (
    FLUID_SBOM_JSON_SCHEMA,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
    IndexedList,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.utils.exceptions import (
    FluidJSONValidationError,
)
import json
from jsonschema import (
    validate,
)
from jsonschema.exceptions import (
    ValidationError,
)
import logging
from pydantic import (
    BaseModel,
)
from typing import (
    Any,
    cast,
)

LOGGER = logging.getLogger(__name__)


class EnumEncoder(json.JSONEncoder):
    def default(self, item: Any) -> Any:  # pylint:disable=arguments-renamed
        if isinstance(item, Enum):
            return item.value
        if isinstance(item, IndexedList):
            return list(item.data)

        if isinstance(item, IndexedDict):
            return dict(item.data)
        if isinstance(item, BaseModel):
            item = item.model_dump if item else None
        if isinstance(item, datetime):
            return item.isoformat()
        return json.JSONEncoder.default(self, item)


def validate_pkgs(sbom_pkgs: list[dict[str, Any]]) -> list[dict[str, Any]]:
    pkgs_to_remove = []
    for index, pkg in enumerate(sbom_pkgs):
        try:
            pkg = json.loads(json.dumps(pkg, cls=EnumEncoder))
            sbom_pkgs[index] = pkg
        except TypeError:
            pkgs_to_remove.append(index)

    for index in reversed(sorted(pkgs_to_remove)):
        sbom_pkgs.pop(index)
    return sbom_pkgs


def format_fluid_sbom(  # pylint:disable=too-many-locals
    packages: list[Package],
    relationships: list[Relationship],
    output: str,
    config: SbomConfig,
) -> None:
    now_utc = datetime.now(UTC).isoformat()
    namespace, version = set_namespace_version(config=config)
    file_path = f"{output}.json"
    sbom_pkgs = [
        {
            "id": package.id_,
            "name": package.name,
            "version": package.version,
            "locations": [
                {
                    "path": location.path(),
                    "line": location.coordinates.line
                    if location.coordinates and location.coordinates.line
                    else None,
                    "layer": location.coordinates.file_system_id
                    if location.coordinates
                    and location.coordinates.file_system_id
                    else None,
                }
                for location in package.locations
            ],
            "licenses": package.licenses,
            "type": package.type.value,
            "language": package.language.value,
            "platform": package.language.get_platform_value(),
            "package_url": package.p_url,
            "found_by": package.found_by,
            "health_metadata": package.health_metadata.model_dump()
            if package.health_metadata
            else None,
            "advisories": [
                advisory.model_dump() for advisory in package.advisories or []
            ],
        }
        for package in packages
    ]

    sbom_pkgs = validate_pkgs(sbom_pkgs)
    sbom_details = {
        "name": namespace,
        "version": version if version else None,
        "timestamp": now_utc,
        "tool": "Fluid-Sbom",
        "organization": "Fluid attacks",
    }
    dependency_map: dict[str, list[str]] = {}

    for relationship in relationships:
        from_pkg = cast(Package, relationship.from_)
        to_pkg = cast(Package, relationship.to_)
        ref = f"{to_pkg.id_}"
        dep = f"{from_pkg.id_}"

        if ref not in dependency_map:
            dependency_map[ref] = []

        dependency_map[ref].append(dep)

    sbom_relationships = [
        {"from": ref, "to": depends_on_list}
        for ref, depends_on_list in dependency_map.items()
    ]
    result = {
        "sbom_details": sbom_details,
        "packages": sbom_pkgs,
        "relationships": sbom_relationships,
    }

    try:
        validate(result, FLUID_SBOM_JSON_SCHEMA)
    except ValidationError as ex:
        raise FluidJSONValidationError(ex) from None

    LOGGER.info(
        "🆗 Valid Fluid JSON format, generating output file at %s", file_path
    )
    with open(file_path, "w", encoding="utf-8") as json_file:
        json.dump(result, json_file, indent=4, cls=EnumEncoder)
    LOGGER.info("✅ Output file successfully generated")



from fnmatch import (
    fnmatch,
)
import logging
from zipfile import (
    BadZipFile,
    ZipFile,
    ZipInfo,
)

LOGGER = logging.getLogger(__name__)


def normalize_zip_entry_name(case_insensitive: bool, entry: str) -> str:
    if case_insensitive:
        entry = entry.lower()
    if not entry.startswith("/"):
        entry = "/" + entry

    return entry


def new_zip_file_manifest(archive_path: str) -> list[ZipInfo]:
    try:
        with ZipFile(archive_path, "r") as myzip:
            return myzip.infolist()
    except BadZipFile:
        LOGGER.warning(
            "Bad zip file",
            extra={
                "extra": {
                    "archive_path": archive_path,
                }
            },
        )
        return []


def zip_glob_match(
    manifest: list[ZipInfo], case_sensitive: bool, *patterns: str
) -> list[str]:
    result = []

    for pattern in patterns:
        for entry in manifest:
            normalized_entry = normalize_zip_entry_name(
                case_sensitive, entry.filename
            )
            if entry.filename.endswith(pattern):
                result.append(entry.filename)
            if case_sensitive:
                pattern = pattern.lower()
            if fnmatch(normalized_entry, pattern):
                result.append(entry.filename)
    return result



from fluid_sbom.internal.cache import (
    dual_cache,
)
from retry_requests import (  # type: ignore
    retry,
)
from typing import (
    NotRequired,
    TypedDict,
)


class NPMPackageAuthor(TypedDict):
    email: str
    name: str


class NPMPackageDist(TypedDict):
    integrity: str
    tarball: str


class NPMPackageLicense(TypedDict):
    type: str
    url: str


class NPMPackageVersion(TypedDict):
    dist: NPMPackageDist
    name: str


class NPMPackageTimeUnpublished(TypedDict):
    time: str
    versions: list[str]


NPMPackage = TypedDict(
    "NPMPackage",
    {
        "author": NotRequired[str | NPMPackageAuthor],
        "dist-tags": NotRequired[dict[str, str]],
        "license": str | NPMPackageLicense,
        "name": str,
        "time": dict[str, str],
        "versions": dict[str, NPMPackageVersion],
    },
)


@dual_cache
def get_npm_package(package_name: str) -> NPMPackage | None:
    response = retry().get(
        f"https://registry.npmjs.org/{package_name}",
        timeout=20,
    )
    if response.status_code != 200:
        return None

    package: NPMPackage = response.json()
    if "unpublished" in package["time"]:
        return None
    return package



from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.config.config import (
    SbomConfig,
)
from fluid_sbom.format.common import (
    process_packages,
)
from fluid_sbom.format.cyclone_dx import (
    format_cyclonedx_sbom,
)
from fluid_sbom.format.fluid import (
    format_fluid_sbom,
)
from fluid_sbom.format.spdx import (
    format_spdx_sbom,
)
from fluid_sbom.pkg.package import (
    Package,
)


def format_sbom(
    packages: list[Package],
    relationships: list[Relationship],
    config: SbomConfig,
) -> None:
    packages = process_packages(packages)
    match config.output_format:
        case "fluid-json":
            format_fluid_sbom(
                packages,
                relationships,
                config.output,
                config,
            )
        case "cyclonedx-json":
            format_cyclonedx_sbom(
                packages,
                relationships,
                config.output_format,
                config.output,
                config,
            )
        case "cyclonedx-xml":
            format_cyclonedx_sbom(
                packages,
                relationships,
                config.output_format,
                config.output,
                config,
            )
        case "spdx-json":
            format_spdx_sbom(
                packages,
                relationships,
                "json",
                config.output,
                config,
            )
        case "spdx-xml":
            format_spdx_sbom(
                packages,
                relationships,
                "xml",
                config.output,
                config,
            )



from bs4 import (
    BeautifulSoup,
    Tag,
)
from contextlib import (
    suppress,
)
from copy import (
    deepcopy,
)
from fluid_sbom import (
    advisories,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.package_information.debian import (
    get_deb_package_version_list,
    get_deb_snapshot,
)
from fluid_sbom.linux.release import (
    Release,
)
from fluid_sbom.pkg.cataloger.debian.parse_copyright import (
    parse_licenses_from_copyright,
)
from fluid_sbom.pkg.cataloger.debian.parse_dpkg_info_files import (
    parse_dpkg_conffile_info,
    parse_dpkg_md5_info,
)
from fluid_sbom.pkg.dpkg import (
    DpkgDBEntry,
    DpkgFileRecord,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Artifact,
    HealthMetadata,
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.pkg.url import (
    purl_qualifiers,
)
from fluid_sbom.utils.file import (
    Digest,
)
import logging
import os
from packageurl import (
    PackageURL,
)
from pathlib import (
    Path,
)
from pydantic import (
    ValidationError,
)
from typing import (
    cast,
    TextIO,
)

LOGGER = logging.getLogger(__name__)


def package_url(pkg: DpkgDBEntry, distro: Release | None = None) -> str:
    qualifiers = {"arch": pkg.architecture}
    if distro and (
        distro.id_ == "debian" or "debian" in (distro.id_like or [])
    ):
        if distro.version_id:
            qualifiers["distro_version_id"] = distro.version_id
        qualifiers["distro_id"] = distro.id_
    if pkg.source:
        qualifiers["upstream"] = (
            f"{pkg.source}@{pkg.source_version}"
            if pkg.source_version
            else pkg.source
        )

    return PackageURL(
        type="deb",
        namespace=distro.id_ if distro and distro.id_ else "",
        name=pkg.package,
        version=pkg.version,
        qualifiers=purl_qualifiers(qualifiers, distro),
        subpath="",
    ).to_string()


def md5_key(metadata: DpkgDBEntry) -> str:
    content_key = metadata.package
    if metadata.architecture not in ("", "all"):
        return f"{content_key}:{metadata.architecture}"
    return content_key


def fetch_md5_content(
    resolver: Resolver, db_location: Location, entry: DpkgDBEntry
) -> tuple[TextIO | None, Location | None] | None:
    if not db_location.coordinates:
        return None
    search_path = str(Path(db_location.coordinates.real_path).parent)
    if not search_path.endswith("status.d"):
        search_path = os.path.join(search_path, "info")
    name = md5_key(entry)
    location = resolver.relative_file_path(
        db_location, os.path.join(search_path, name + ".md5sums")
    )
    if not location:
        location = resolver.relative_file_path(
            db_location, os.path.join(search_path, entry.package + ".md5sums")
        )
    if not location:
        return None

    reader = resolver.file_contents_by_location(location)
    if not reader:
        LOGGER.warning(
            "failed to fetch deb md5 contents (package=%s)",
            entry.package,
        )
    return reader, location


def fetch_conffile_contents(
    resolver: Resolver, db_location: Location, entry: DpkgDBEntry
) -> tuple[TextIO | None, Location | None] | None:
    if not db_location.coordinates:
        return None
    parent_path = str(Path(db_location.coordinates.real_path).parent)

    name = md5_key(entry)
    location = resolver.relative_file_path(
        db_location, os.path.join(parent_path, "info", name + ".conffiles")
    )
    if not location:
        location = resolver.relative_file_path(
            db_location,
            os.path.join(parent_path, "info", entry.package + ".conffiles"),
        )
    if not location:
        return None, None
    reader = resolver.file_contents_by_location(location)
    if not reader:
        LOGGER.warning(
            "failed to fetch deb conffiles contents (package=%s)",
            entry.package,
        )
    return reader, location


def get_additional_file_listing(
    resolver: Resolver, db_location: Location, entry: DpkgDBEntry
) -> tuple[list[DpkgFileRecord], list[Location]]:
    files: list[DpkgFileRecord] = []
    locations: list[Location] = []
    md5_result = fetch_md5_content(resolver, db_location, entry)
    if not md5_result:
        return files, locations
    md5_reader, md5_location = md5_result
    if md5_reader is not None and md5_location is not None:
        files.extend(parse_dpkg_md5_info(md5_reader))
        locations.append(md5_location)
    conffiles = fetch_conffile_contents(resolver, db_location, entry)
    if not conffiles:
        return files, locations
    conffiles_reader, conffiles_location = conffiles

    if conffiles_reader is not None and conffiles_location is not None:
        files.extend(parse_dpkg_conffile_info(conffiles_reader))
        locations.append(conffiles_location)

    return files, locations


def merge_file_listing(
    resolver: Resolver, db_location: Location, pkg: Package
) -> None:
    metadata: DpkgDBEntry = cast(DpkgDBEntry, pkg.metadata)
    files, info_locations = get_additional_file_listing(
        resolver, db_location, metadata
    )
    for new_file in files:
        exists = False
        for existing_file in metadata.files or []:
            if existing_file.path == new_file.path:
                exists = True
                break
        if not exists and metadata.files:
            metadata.files.append(new_file)
    sorted_files = sorted(metadata.files or [], key=lambda x: x.path)
    metadata.files = sorted_files
    pkg.metadata = metadata
    pkg.locations.extend(info_locations)


def fetch_copyright_contents(
    resolver: Resolver | None, db_location: Location, metadata: DpkgDBEntry
) -> tuple[TextIO | None, Location | None]:
    if not resolver:
        return None, None

    copyright_path = os.path.join(
        "/usr/share/doc", metadata.package, "copyright"
    )
    location = resolver.relative_file_path(db_location, copyright_path)

    if not location:
        return None, None

    reader = resolver.file_contents_by_location(location)
    if not reader:
        LOGGER.warning(
            "failed to fetch deb copyright contents (package=%s)",
            metadata.package,
        )

    return reader, location


def add_licenses(
    resolver: Resolver, db_location: Location, pkg: Package
) -> None:
    metadata: DpkgDBEntry = cast(DpkgDBEntry, pkg.metadata)

    pkg.licenses = []
    copyright_reader, copyright_location = fetch_copyright_contents(
        resolver, db_location, metadata
    )

    if copyright_reader is not None and copyright_location is not None:
        licenses_strs = parse_licenses_from_copyright(copyright_reader)
        pkg.licenses = licenses_strs


def new_dpkg_package(
    entry: DpkgDBEntry,
    db_location: Location,
    _resolver: Resolver | None,
    release: Release | None = None,
) -> Package | tuple[Package, Package] | None:
    try:
        dpkg = Package(
            name=entry.package,
            version=entry.version or "",
            licenses=[],
            p_url=package_url(entry, release),
            locations=[db_location],
            type=PackageType.DebPkg,
            metadata=entry,
            found_by=None,
            language=Language.UNKNOWN_LANGUAGE,
        )
        if _resolver is not None:
            # side effects
            merge_file_listing(_resolver, db_location, dpkg)
            add_licenses(_resolver, db_location, dpkg)

        source_dpkg: Package | None = None
        if (entry.source and entry.source != dpkg.name) or (
            entry.source
            and entry.source_version
            and entry.source_version != dpkg.version
        ):
            new_entry = deepcopy(entry)
            new_entry.package = entry.source
            new_entry.version = entry.source_version or dpkg.version
            new_entry.source = None
            new_entry.source_version = None
            new_entry.dependencies = None
            new_entry.pre_dependencies = None

            source_dpkg = deepcopy(dpkg)
            source_dpkg.name = new_entry.package
            source_dpkg.version = new_entry.version
            source_dpkg.p_url = package_url(new_entry, release)

        if source_dpkg:
            return dpkg, source_dpkg
        return dpkg
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types "
            "are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": db_location.path(),
                }
            },
        )
        return None


def _search_download_url(
    package: Package, arch: str | None = None
) -> tuple[str, str | None] | None:
    html_download = get_deb_snapshot(package.name, package.version)
    if not html_download:
        return None
    parsed = BeautifulSoup(html_download, features="html.parser")
    tags: list[Tag] = parsed.find_all("a", href=True)
    tag_href: str | list[str] | None = None
    for tag in tags:
        if (
            tag.text.endswith(".deb")
            and package.name in tag.text
            and package.version in tag.text
            and (arch in tag.text if arch else True)
        ):
            tag_href = tag.get("href")
            break
    else:
        for tag in tags:
            if (
                tag.text.endswith(".deb")
                and package.name in tag.text
                and package.version in tag.text
            ):
                tag_href = tag.get("href")
                break
    if not tag_href:
        return None
    sha1_hash: str | None = None
    with suppress(IndexError):
        sha1_hash = list(tag.fetchPrevious("code", limit=1))[0].text

    return f"https://snapshot.debian.org{tag_href}", sha1_hash


def complete_package(
    package: Package, release: Release | None = None, arch: str | None = None
) -> Package:
    if release:
        pkg_advisories = advisories.get_package_advisories(
            package, distro_id=release.id_, distro_version=release.version_id
        )
    else:
        pkg_advisories = advisories.get_package_advisories(package)
    if pkg_advisories:
        package.advisories = pkg_advisories
    version_code_name: str | None = None
    if release:
        version_code_name = release.version_code_name
    versions_list = get_deb_package_version_list(
        package.name, version_code_name
    )
    package.health_metadata = HealthMetadata()
    if versions_list:
        package.health_metadata = HealthMetadata(
            latest_version=versions_list[0]["version"],
        )
    if package.metadata and hasattr(package.metadata, "maintainer"):
        package.health_metadata.authors = package.metadata.maintainer
    download_url_item = _search_download_url(package, arch)
    if download_url_item:
        digest_value = download_url_item[1] or None
        package.health_metadata.artifact = Artifact(
            url=download_url_item[0],
            integrity=Digest(
                algorithm="sha1" if digest_value else None, value=digest_value
            ),
        )
    return package



from contextlib import (
    suppress,
)
import diskcache
from diskcache import (
    Lock,
)
import diskcache.core
import functools
import hashlib
import os
import pickle
from platformdirs import (
    user_cache_dir,
)
import redis
import socket
from typing import (
    Any,
    Callable,
    cast,
    TypeVar,
)

REDIS_CACHE_ENDPOINT = os.environ.get("CACHE_URL")
REDIS_CLIENT = None

TVar = TypeVar("TVar")  # pylint: disable=invalid-name
TFun = TypeVar(  # pylint: disable=invalid-name
    "TFun", bound=Callable[..., Any]
)

if REDIS_CACHE_ENDPOINT:
    try:
        host, _port = REDIS_CACHE_ENDPOINT.split(":", maxsplit=1)
        port = int(_port)

        # Intentar conectarse al endpoint de Redis
        socket.create_connection((host, port), timeout=2)

        REDIS_CLIENT = redis.StrictRedis(
            host=host,
            port=port,
            ssl=True,
            username="readwrite",
            password=os.environ.get("CACHE_USER_WRITE_PASSWORD"),
        )
    except (socket.error, ValueError):
        REDIS_CLIENT = None

DISK_CACHE = diskcache.Cache(user_cache_dir("fluid-sbom", "fluidattacks"))


# 4 weeks
TTL = 604800 * 4


def make_hashable(item: Any) -> str:
    serialized_object = pickle.dumps(item)

    return hashlib.sha256(serialized_object).hexdigest()


def generate_cache_key(func: Any, args: Any, kwargs: dict[Any, Any]) -> str:
    key = f"{func.__module__}.{func.__name__}-"
    key += str(make_hashable((args, kwargs)))
    return hashlib.sha256(key.encode()).hexdigest()


def dual_cache(func: TVar) -> TVar:
    _func = cast(Callable[..., Any], func)

    @functools.wraps(_func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        cache_key = generate_cache_key(_func, args, kwargs)

        # Try to recover from disk cache
        if value := DISK_CACHE.get(cache_key):
            return value
        # Try to recover from Redis cache (ElastiCache) if available
        if REDIS_CLIENT:
            cached_result = REDIS_CLIENT.get(cache_key)

            if cached_result:
                result = pickle.loads(cached_result)  # type: ignore
                with Lock(DISK_CACHE, "dual_cache"):
                    DISK_CACHE.set(
                        cache_key, result, expire=TTL, retry=True
                    )  # Save to disk with TTL

                return result

        # Run the function and store the result in both caches
        result = _func(*args, **kwargs)
        with suppress(diskcache.core.Timeout):
            with Lock(DISK_CACHE, "dual_cache"):
                DISK_CACHE.set(
                    cache_key, result, expire=TTL, retry=True
                )  # Save to disk with TTL
        if REDIS_CLIENT:
            REDIS_CLIENT.setex(
                cache_key, TTL, pickle.dumps(result)
            )  # Save to Redis with TTL
        return result

    return cast(TVar, wrapper)



from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection import (
    toml,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.python.package import (
    package_url,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.python import (
    PythonPackage,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
from pydantic import (
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def parse_pipfile_deps(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    file_content = reader.read_closer.read()
    toml_content = toml.parse_toml_with_tree_sitter(file_content)
    toml_packages = toml_content.get("packages", {})
    for package in toml_packages:
        version = toml_packages[package]
        if not isinstance(version, str):
            version = version.get("version", "*")
        if "*" in version:
            continue

        version = version.strip("=<>~^ ")
        try:
            packages.append(
                Package(
                    name=package,
                    version=version,
                    locations=[
                        (
                            _get_location(
                                reader.location, package.position.start.line
                            )
                            if isinstance(package, IndexedDict)
                            else reader.location
                        )
                    ],
                    language=Language.PYTHON,
                    type=PackageType.PythonPkg,
                    metadata=PythonPackage(
                        name=package,
                        version=version,
                    ),
                    p_url=package_url(
                        name=package, version=version, package=None
                    ),
                    licenses=[],
                )
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types "
                "are incorrect.",
                extra={
                    "extra": {
                        "exception": ex.errors(include_url=False),
                        "location": reader.location.path(),
                    }
                },
            )
            continue
    return packages, []



from fluid_sbom.artifact.relationship import (
    Relationship,
    RelationshipType,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.linux.release import (
    Release,
)
from fluid_sbom.pkg.apk import (
    ApkDBEntry,
    ApkFileRecord,
)
from fluid_sbom.pkg.cataloger.alpine.package import (
    new_package,
    ParsedData,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.utils.file import (
    Digest,
)
import os
from pydantic import (
    BaseModel,
    ConfigDict,
)
import re
from typing import (
    cast,
    TypedDict,
)

APK_DB_GLOB = "**/lib/apk/db/installed"
FileInfo = TypedDict("FileInfo", {"uid": str, "gid": str, "perms": str})


class ApkField(BaseModel):
    name: str
    value: str
    model_config = ConfigDict(frozen=True)


class ApkFileParsingContext(BaseModel):
    files: list[ApkFileRecord]
    index_of_latest_directory: int
    index_of_latest_regular_file: int


def _parse_list_values(
    value: str | None, delimiter: str | None = None
) -> list[str]:
    delimiter = delimiter or " "
    if not value:
        return []
    return value.split(delimiter)


def _process_file_info(
    info: str,
) -> FileInfo | None:
    file_info = info.split(":")
    if len(file_info) < 3:
        return None
    return {"uid": file_info[0], "gid": file_info[1], "perms": file_info[2]}


def process_checksum(value: str) -> Digest:
    algorithm = "md5"
    if value.startswith("Q1"):
        algorithm = "'Q1'+base64(sha1)"
    return Digest(algorithm=algorithm, value=value)


def find_release(db_path: str) -> list[Release] | None:
    location = os.path.normpath(
        os.path.join(db_path, "../../../etc/apk/repositories")
    )
    if not os.path.exists(location):
        return None
    return None


def parse_apk_db(
    _resolver: Resolver,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]] | None:
    content = reader.read_closer.read()
    apks = [
        parsed_package
        for package in content.strip().split("\n\n")
        if package and (parsed_package := parse_package(package)) is not None
    ]

    entries = [
        entry
        for apk in apks
        if (
            entry := new_package(
                apk, _env.linux_release if _env else None, reader.location
            )
        )
        is not None
    ]

    return entries, discover_package_dependencies(entries)


def parse_package(package: str) -> ParsedData | None:
    data: dict[str, str] = {}
    ctx = ApkFileParsingContext(
        files=[], index_of_latest_directory=-1, index_of_latest_regular_file=-1
    )
    lines = package.split("\n")
    key = ""
    for line in lines:
        key = process_line(line, key, data, ctx)
    return construct_apk(data, ctx)


def process_line(
    line: str, key: str, data: dict[str, str], ctx: ApkFileParsingContext
) -> str:
    if ":" in line:
        key, value = line.split(":", 1)
        data[key] = value
        update_context_with_line(key, value, ctx)
    elif key and key in data:
        data[key] += "\n" + line.strip()
    return key


def update_context_with_line(
    key: str, value: str, ctx: ApkFileParsingContext
) -> None:
    match key:
        case "F":
            ctx.files.append(ApkFileRecord(path=os.path.join("/", value)))
            ctx.index_of_latest_directory = len(ctx.files) - 1
        case "M":
            index = ctx.index_of_latest_directory
            latest = ctx.files[index]
            if file_info := _process_file_info(value):
                latest.owner_uid = file_info["uid"]
                latest.owner_gid = file_info["gid"]
                latest.permissions = file_info["perms"]
            ctx.files[index] = latest
        case "R":
            index = ctx.index_of_latest_directory
            if index < 0:
                regular_file = os.path.join("/", value)
            else:
                latest_dir_path = ctx.files[index].path
                regular_file = os.path.join(latest_dir_path, value)
            ctx.files.append(ApkFileRecord(path=regular_file))
            ctx.index_of_latest_regular_file = len(ctx.files) - 1
        case "a":
            index = ctx.index_of_latest_regular_file
            latest = ctx.files[index]
            if file_info := _process_file_info(value):
                latest.owner_uid = file_info["uid"]
                latest.owner_gid = file_info["gid"]
                latest.permissions = file_info["perms"]
            ctx.files[index] = latest
        case "Z":
            index = ctx.index_of_latest_regular_file
            latest = ctx.files[index]
            latest.digest = process_checksum(value)


def construct_apk(
    data: dict[str, str], ctx: ApkFileParsingContext
) -> ParsedData | None:
    if not (package := data.get("P")) or not (version := data.get("V")):
        return None

    return ParsedData(
        apk_db_entry=ApkDBEntry(
            package=package,
            origin_package=data.get("o"),
            maintainer=data.get("m"),
            version=version,
            architecture=data.get("A"),
            url=data.get("U", ""),
            description=data.get("T", ""),
            size=data.get("S", ""),
            installed_size=data.get("I"),
            dependencies=_parse_list_values(data.get("D")),
            provides=_parse_list_values(data.get("p")),
            checksum=data.get("C"),
            git_commit=data.get("c"),
            files=ctx.files,
        ),
        license=data.get("L"),
    )


def _build_lookup_table(pkgs: list[Package]) -> dict[str, list[Package]]:
    lookup: dict[str, list[Package]] = {}

    for pkg in pkgs:
        apkg: ApkDBEntry = cast(ApkDBEntry, pkg.metadata)
        if pkg.name not in lookup:
            lookup[pkg.name] = [pkg]
        else:
            lookup[pkg.name].append(pkg)

        for provides in apkg.provides:
            provides_k = strip_version_specifier(provides)
            if provides_k not in lookup:
                lookup[provides_k] = [pkg]
            else:
                lookup[provides_k].append(pkg)

    return lookup


def discover_package_dependencies(  # NOSONAR
    pkgs: list[Package],
) -> list[Relationship]:
    lookup: dict[str, list[Package]] = _build_lookup_table(pkgs)
    relationships: list[Relationship] = []

    for pkg in pkgs:
        apkg = cast(ApkDBEntry, pkg.metadata)
        for dep_specifier in apkg.dependencies:
            dep = strip_version_specifier(dep_specifier)
            for dep_pk in lookup.get(dep, []):
                relationships.append(
                    Relationship(
                        from_=dep_pk,
                        to_=pkg,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        data={},
                    )
                )
    return relationships


def strip_version_specifier(version: str) -> str:
    return re.split("[<>=]", version)[0]



from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.collection.types import (
    IndexedDict,
)
from fluid_sbom.internal.collection.yaml import (
    parse_yaml_with_tree_sitter,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from packageurl import (
    PackageURL,
)
from typing import (
    cast,
    ItemsView,
)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def _get_packages(
    reader: LocationReadCloser,
    dependencies: IndexedDict | None,
) -> list[Package]:
    if dependencies is None:
        return []

    general_location = _get_location(
        reader.location, dependencies.position.start.line
    )
    items: ItemsView[str, IndexedDict | str] = dependencies.items()
    return [
        Package(
            name=name,
            version=version,
            locations=[general_location],
            language=Language.DART,
            licenses=[],
            type=PackageType.DartPubPkg,
            p_url=PackageURL(  # type: ignore
                type="pub",
                name=name,
                version=version,
            ).to_string(),
        )
        for name, version in items
        if isinstance(version, str)
    ]


def parse_pubspec_yaml(
    _: Resolver | None, __: Environment | None, reader: LocationReadCloser
) -> tuple[list[Package], list[Relationship]]:
    content = cast(
        IndexedDict, parse_yaml_with_tree_sitter(reader.read_closer.read())
    )
    deps: IndexedDict | None = content.get("dependencies")
    dev_deps: IndexedDict | None = content.get("dev_dependencies")
    packages = [
        *_get_packages(reader, deps),
        *_get_packages(reader, dev_deps),
    ]
    return packages, []



# pylint:disable=invalid-name
from enum import (
    Enum,
)


class PackageType(Enum):
    UnknownPkg = "UnknownPackage"
    AlpmPkg = "alpm"
    ApkPkg = "apk"
    BinaryPkg = "binary"
    CocoapodsPkg = "pod"
    ConanPkg = "conan"
    DartPubPkg = "dart-pub"
    DebPkg = "deb"
    DotnetPkg = "dotnet"
    ErlangOTPPkg = "erlang-otp"
    GemPkg = "gem"
    GithubActionPkg = "github-action"
    GithubActionWorkflowPkg = "github-action-workflow"
    GoModulePkg = "go-module"
    GraalVMNativeImagePkg = "graalvm-native-image"
    HackagePkg = "hackage"
    HexPkg = "hex"
    JavaPkg = "java-archive"
    JenkinsPluginPkg = "jenkins-plugin"
    KbPkg = "msrc-kb"
    LinuxKernelPkg = "linux-kernel"
    LinuxKernelModulePkg = "linux-kernel-module"
    NixPkg = "nix"
    NpmPkg = "npm"
    PhpComposerPkg = "php-composer"
    PhpPeclPkg = "php-pecl-pkg"
    PortagePkg = "portage"
    PythonPkg = "python"
    Rpkg = "R-package"
    RpmPkg = "rpm"
    RustPkg = "rust-crate"
    SwiftPkg = "swift"
    WordpressPluginPkg = "wordpress-plugin"



from enum import (
    Enum,
)
from pydantic import (
    BaseModel,
    ConfigDict,
)
from typing import (
    Any,
    Hashable,
)


class RelationshipType(Enum):
    OWNERSHIP_BY_FILE_OVERLAP_RELATIONSHIP: str = "ownership-by-file-overlap"
    EVIDENT_BY_RELATIONSHIP: str = "evident-by"
    CONTAINS_RELATIONSHIP: str = "contains"
    DEPENDENCY_OF_RELATIONSHIP: str = "dependency-of"
    DESCRIBED_BY_RELATIONSHIP: str = "described-by"


class Relationship(BaseModel):
    from_: Hashable
    to_: Hashable
    type: RelationshipType
    data: Any
    model_config = ConfigDict(frozen=True)



from contextlib import (
    suppress,
)
from fluid_sbom.context.image import (
    ImageContext,
)
from fluid_sbom.file.location import (
    Location,
    new_location_from_image,
)
from fluid_sbom.file.metadata import (
    Metadata,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.file.type import (
    get_type_from_tar_member,
)
from fluid_sbom.sources.docker import (
    ImageMetadata,
)
import fnmatch
import os
from pathlib import (
    Path,
)
from pydantic import (
    BaseModel,
)
import tarfile
import tempfile
from typing import (
    Callable,
    Generator,
    TextIO,
)


class ContainerImage(Resolver, BaseModel):
    img: ImageMetadata
    context: ImageContext
    lazy: bool = True

    def _has_path_lazy(self, path: str) -> bool:
        success = []
        paths = [path, path.lstrip("/")]
        for layer_info in self.context.manifest["Layers"]:
            layer_dir = os.path.join(layer_info["id"], "layer.tar")
            with tarfile.open(
                os.path.join(self.context.layers_dir, layer_dir)
            ) as tar:
                exists = False
                for _path in paths:
                    try:
                        tar.getmember(_path)
                        success.append(True)
                        exists = True
                        break
                    except KeyError:
                        success.append(False)
                if exists:
                    break

        return any(success)

    def _has_path(self, path: str) -> bool:
        layer_ids = [x["digest"] for x in self.context.manifest["layers"]]

        success = []
        path = path.lstrip(os.path.sep)
        for layer_id in layer_ids:
            p_file_path = os.path.join(
                self.context.full_extraction_dir, layer_id, path
            )
            if os.path.exists(p_file_path):
                success.append(True)
                break

        return any(success)

    def has_path(self, path: str) -> bool:
        if self.lazy:
            return self._has_path_lazy(path)

        return self._has_path(path)

    def _search_path_lazy(self, path: str) -> list[Location]:
        # A path can be in multiple layers
        locations: list[Location] = []
        for layer_info in self.context.manifest["layers"]:
            layer_dir = os.path.join(layer_info["id"], "layer.tar")
            with tarfile.open(
                os.path.join(self.context.layers_dir, layer_dir)
            ) as tar:
                with suppress(KeyError):
                    member = tar.getmember(path)
                    locations.append(
                        new_location_from_image(
                            member.name,
                            str(Path(layer_dir).parent),
                        )
                    )

        return locations

    def _search_path(self, path: str) -> list[Location]:
        if self.lazy:
            return self._search_path_lazy(path)

        locations: list[Location] = []
        layer_ids = [x["digest"] for x in self.context.manifest["layers"]]

        # A path can be in multiple layers
        for layer_id in layer_ids:
            p_file_path = os.path.join(
                self.context.full_extraction_dir,
                layer_id,
                path.lstrip(os.path.sep),
            )
            if os.path.exists(p_file_path):
                locations.append(
                    new_location_from_image(
                        path,
                        layer_id,
                        p_file_path,
                    )
                )
        return locations

    def files_by_path(self, *paths: str) -> list[Location]:
        locations: list[Location] = []

        for path in paths:
            if find_path := self._search_path(path.lstrip("/")):
                locations.extend(find_path)
            else:
                locations.extend(self._search_path(path))
        return locations

    def files_by_glob(self, *patterns: str) -> list[Location]:
        locations: list[Location] = []
        for layer_info in self.context.manifest["Layers"]:
            layer_dir = os.path.join(layer_info["id"], "layer.tar")
            with tarfile.open(
                os.path.join(self.context.layers_dir, layer_dir)
            ) as tar:
                for member in tar.getmembers():
                    for glob in patterns:
                        if fnmatch.fnmatch(member.name, glob):
                            locations.append(
                                new_location_from_image(
                                    member.name,
                                    str(Path(layer_dir).parent),
                                    os.path.join(
                                        self.context.full_extraction_dir,
                                        member.name,
                                    ),
                                )
                            )
        return locations

    def _file_contents_by_location_lazy(
        self,
        location: Location,
        *,
        function_reader: Callable[..., TextIO] | None = None,
        mode: str | None = None,
    ) -> TextIO | None:
        for layer_info in self.context.manifest["Layers"]:
            layer_dir = os.path.join(layer_info["id"], "layer.tar")
            if (
                location.coordinates
                and location.coordinates.file_system_id
                and location.coordinates.file_system_id not in layer_dir
            ):
                continue
            with tarfile.open(
                os.path.join(self.context.layers_dir, layer_dir)
            ) as tar:
                with suppress(KeyError):
                    if location.access_path:
                        member = tar.getmember(location.access_path)
                        if member.linkpath:
                            link_target_path = os.path.normpath(
                                os.path.join(
                                    os.path.dirname(member.name),
                                    member.linkname,
                                )
                            )
                            member = tar.getmember(link_target_path)
                        temp_dir = tempfile.mkdtemp()
                        tar.extract(tar.getmember(member.name), temp_dir)
                        return (function_reader or open)(  # type: ignore
                            os.path.join(temp_dir, member.name),
                            encoding="utf-8",
                            mode=mode or "r",
                        )

        return None

    def file_contents_by_location(
        self,
        location: Location,
        *,
        function_reader: Callable[..., TextIO] | None = None,
        mode: str | None = None,
    ) -> TextIO | None:
        if (
            not self.lazy
            and location.coordinates
            and os.path.exists(location.coordinates.real_path)
        ):
            return (function_reader or open)(  # type: ignore
                location.coordinates.real_path,
                encoding="utf-8",
                mode=mode or "r",
            )

        if self.lazy:
            return self._file_contents_by_location_lazy(
                location, function_reader=function_reader, mode=mode
            )

        return None

    def file_metadata_by_location(
        self,
        location: Location,
    ) -> Metadata | None:
        if (
            location.coordinates is not None
            and location.coordinates.file_system_id is not None
            and (
                _layer_info := self.context.get_layer_info(
                    location.coordinates.file_system_id
                )
            )
        ):
            with tarfile.open(os.path.join(_layer_info.tar_full_path)) as tar:
                tar_member = tar.getmember(location.coordinates.real_path)

                return Metadata(
                    path=tar_member.name,
                    link_destination=tar_member.linkname,
                    user_id=tar_member.uid,
                    group_id=tar_member.gid,
                    type=get_type_from_tar_member(tar_member),
                    mime_type="",
                )

        return None

    def files_by_mime_type(self, _mime_type: str) -> list[Location]:
        raise NotImplementedError

    def relative_file_path(self, _: Location, _path: str) -> Location | None:
        files = self.files_by_path(_path)
        if not files:
            return None
        return files[0]

    def walk_file_lazy(self) -> Generator[str, None, None]:
        for layer_info in self.context.manifest["Layers"]:
            layer_dir = os.path.join(layer_info["id"], "layer.tar")
            with tarfile.open(
                os.path.join(self.context.layers_dir, layer_dir)
            ) as tar:
                for member in tar.getmembers():
                    yield member.name

    def walk_file(self) -> Generator[str, None, None]:
        if self.lazy:
            yield from self.walk_file_lazy()
        else:
            layer_ids = [x["digest"] for x in self.context.manifest["layers"]]
            for layer_id in layer_ids:
                current_layer_path = os.path.join(
                    self.context.full_extraction_dir, layer_id
                )
                for root, _, files in os.walk(current_layer_path):
                    for file in files:
                        yield os.path.join(
                            os.path.sep,
                            root.replace(current_layer_path, ""),
                            file,
                        )



from fluid_sbom.file.coordinates import (
    Coordinates,
)
from pydantic import (
    BaseModel,
)


class LocationMetadata(BaseModel):
    annotations: dict[str, str]

    def merge(self, other: "LocationMetadata") -> "LocationMetadata":
        return LocationMetadata(
            annotations={**self.annotations, **other.annotations}
        )


class LocationData(BaseModel):
    coordinates: Coordinates
    access_path: str

    def __hash__(self) -> int:
        return hash(self.access_path) + hash(self.coordinates.file_system_id)


class Location(BaseModel):
    coordinates: Coordinates | None = None
    access_path: str | None = None
    annotations: dict[str, str] | None = None

    def with_annotation(self, key: str, value: str) -> "Location":
        if not self.annotations:
            self.annotations = {}
        self.annotations[key] = value
        return self

    def path(self) -> str:
        return (
            self.access_path
            or (self.coordinates.real_path if self.coordinates else "")
            or ""
        )


def new_location_from_image(
    access_path: str | None, layer_id: str, real_path: str | None = None
) -> Location:
    if access_path and not access_path.startswith("/"):
        access_path = f"/{access_path}"
    return Location(
        coordinates=Coordinates(
            real_path=real_path or "", file_system_id=layer_id
        ),
        access_path=access_path,
        annotations={},
    )


def new_location(real_path: str) -> Location:
    return Location(
        coordinates=Coordinates(
            real_path=real_path,
        ),
        access_path=real_path,
        annotations={},
    )



from fluid_sbom.artifact.relationship import (
    Relationship,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.location_read_closer import (
    LocationReadCloser,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.pkg.cataloger.generic.parser import (
    Environment,
)
from fluid_sbom.pkg.cataloger.python.package import (
    new_package_for_package,
)
from fluid_sbom.pkg.cataloger.python.parse_wheel_egg_metadata import (
    parse_wheel_or_egg_metadata,
    ParsedData,
)
from fluid_sbom.pkg.cataloger.python.parse_wheel_egg_record import (
    parse_installed_files,
    parse_wheel_or_egg_record,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.python import (
    PythonDirectURLOriginInfo,
    PythonFileRecord,
)
import json
import os
from pathlib import (
    Path,
)


def fetch_record_files(
    resolver: Resolver, metadata_location: Location
) -> tuple[list[PythonFileRecord] | None, list[Location] | None]:
    files: list[PythonFileRecord] = []
    sources: list[Location] = []
    if not metadata_location.coordinates:
        return None, None

    record_path = os.path.join(
        Path(metadata_location.coordinates.real_path).parent, "RECORD"
    )
    record_ref: Location | None = resolver.relative_file_path(
        metadata_location, record_path
    )
    if not record_ref:
        return None, None
    sources.append(record_ref)
    record_content = resolver.file_contents_by_location(record_ref)
    if not record_content:
        return None, None
    records = parse_wheel_or_egg_record(record_content)
    files.extend(records)
    return files, sources


def fetch_installed_packages(
    resolver: Resolver,
    metadata_location: Location,
    site_packages_root_path: str,
) -> tuple[list[PythonFileRecord] | None, list[Location] | None]:
    files: list[PythonFileRecord] = []
    sources: list[Location] = []

    if not metadata_location.coordinates:
        return None, None

    installed_files_path = os.path.join(
        metadata_location.coordinates.real_path, "installed-files.txt"
    )
    installed_files_ref = resolver.relative_file_path(
        metadata_location, installed_files_path
    )

    if installed_files_ref:
        sources.append(installed_files_ref)
        installed_files_content = resolver.file_contents_by_location(
            installed_files_ref
        )
        if not installed_files_content:
            return None, None

        installed_files = parse_installed_files(
            installed_files_content,
            installed_files_path,
            site_packages_root_path,
        )

        files.extend(installed_files)
    return files, sources


def fetch_top_level_packages(
    resolver: Resolver, metadata_location: Location
) -> tuple[list[str] | None, list[Location] | None]:
    pkgs: list[str] = []
    if not metadata_location.coordinates:
        return None, None
    parent_dir = str(Path(metadata_location.coordinates.real_path).parent)
    top_level_path = os.path.join(parent_dir, "top_level.txt")
    top_level_location = resolver.relative_file_path(
        metadata_location, top_level_path
    )
    if not top_level_location:
        return None, None

    sources = [top_level_location]
    top_level_content = resolver.file_contents_by_location(top_level_location)
    if not top_level_content:
        return None, None

    for line in top_level_content.readlines():
        pkgs.append(line.rstrip("\n"))

    return pkgs, sources


def fetch_direct_url_data(
    resolver: Resolver, metadata_location: Location
) -> tuple[PythonDirectURLOriginInfo | None, list[Location] | None]:
    if not metadata_location.coordinates:
        return None, None
    direct_url_path = os.path.join(
        metadata_location.coordinates.real_path, "direct_url.json"
    )
    direct_url_location = resolver.relative_file_path(
        metadata_location, direct_url_path
    )
    if not direct_url_location:
        return None, None

    sources = [direct_url_location]
    direct_url_content = resolver.file_contents_by_location(
        direct_url_location
    )
    if not direct_url_content:
        return None, None
    decoded_dat = json.load(direct_url_content)
    return (
        PythonDirectURLOriginInfo(
            url=decoded_dat["url"],
            commit_id=decoded_dat["vsc_info"]["commit_id"],
            vcs=decoded_dat["vsc_info"]["vsc"],
        ),
        sources,
    )


def assemble_egg_or_wheel_metadata(
    resolver: Resolver, metadata_location: Location
) -> tuple[ParsedData | None, list[Location] | None]:
    sources_r = [metadata_location]
    sources: list[Location] | None = None
    metadata_content = resolver.file_contents_by_location(metadata_location)
    if not metadata_content or not metadata_location.coordinates:
        return None, None
    p_data = parse_wheel_or_egg_metadata(
        metadata_location.coordinates.real_path, metadata_content
    )
    if not p_data:
        return None, None
    records, sources = fetch_record_files(resolver, metadata_location)
    if not records or not sources:
        records, sources = fetch_installed_packages(
            resolver,
            metadata_location,
            p_data.python_package.site_package_root_path or "",
        )
    if sources:
        sources_r.extend(sources or [])
    p_data.python_package.files = records

    top_packages, sources = fetch_top_level_packages(
        resolver, metadata_location
    )
    sources_r.extend(sources or [])
    p_data.python_package.top_level_packages = top_packages
    direct_url, sources = fetch_direct_url_data(resolver, metadata_location)

    if direct_url and sources_r:
        sources_r.extend(sources or [])
        p_data.python_package.direct_url_origin = direct_url

    return p_data, sources_r


def parse_wheel_or_egg(
    resolver: Resolver, _: Environment, reader: LocationReadCloser
) -> tuple[list[Package], list[Relationship]]:
    p_data, _sources = assemble_egg_or_wheel_metadata(
        resolver, reader.location
    )
    if not p_data:
        return [], []

    pkg = new_package_for_package(resolver, p_data, reader.location)

    return [pkg] if pkg is not None else [], []



from fluid_sbom.internal.package_information.rust import (
    CRATES_ENDPOINT,
    get_cargo_package,
    Version,
)
from fluid_sbom.pkg.package import (
    Artifact,
    HealthMetadata,
    Package,
)
from fluid_sbom.utils.file import (
    Digest,
)
from packageurl import (
    PackageURL,
)


def _set_health_metadata_artifact(
    package: Package,
    current_package: Version | None,
) -> Package:
    if current_package and package.health_metadata:
        digest_value = current_package.get("checksum")
        package.health_metadata.artifact = Artifact(
            url=f"{CRATES_ENDPOINT}{current_package['dl_path']}",
            integrity=Digest(
                algorithm="sha256" if digest_value else None,
                value=digest_value,
            ),
        )
    return package


def complete_package(package: Package) -> Package:
    cargo_package = get_cargo_package(package.name)
    if not cargo_package:
        return package
    current_package = next(
        (
            version
            for version in cargo_package["versions"]
            if version["num"] == package.version
        ),
        None,
    )
    try:
        package.health_metadata = HealthMetadata(
            latest_version=cargo_package["crate"]["max_stable_version"],
            latest_version_created_at=(cargo_package["crate"]["updated_at"]),
        )
    except IndexError:
        return package
    package = _set_health_metadata_artifact(package, current_package)
    if (
        package.health_metadata
        and cargo_package["versions"][0]["published_by"]
    ):
        package.health_metadata.authors = cargo_package["versions"][0][
            "published_by"
        ]["name"]
    if (
        not package.licenses
        and (licenses := cargo_package["versions"][0].get("license"))
        and isinstance(licenses, str)
    ):
        package.licenses = [cargo_package["versions"][0]["license"]]

    return package


def package_url(name: str, version: str) -> str:
    return PackageURL(  # type: ignore
        type="cargo",
        namespace="",
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()



from fluid_sbom.internal.collection.types import (
    IndexedDict,
    IndexedList,
)
from fluid_sbom.utils.exceptions import (
    UnexpectedNode,
)
import logging
import os
from tree_sitter import (
    Language,
    Node,
    Parser,
)
from typing import (
    Any,
)

LOGGER = logging.getLogger(__name__)


def _handle_array_node(node: Node) -> tuple[Node, IndexedList]:
    data = IndexedList(node)
    for child in node.children:
        if child.type not in ("[", "]", ","):
            value_node, value = handle_json_node(child)
            data.append((value, value_node))
    return node, data


def _handle_object_node(node: Node) -> tuple[Node, IndexedDict]:
    data = IndexedDict(node)
    for child in node.children:
        if child.type == "pair":
            key_n, _, value_n = child.children
            key = key_n.text[1:-1].decode("utf-8")
            value_node, value_value = handle_json_node(value_n)
            data[(key, key_n)] = (value_value, value_node)
    return node, data


def handle_json_node(node: Node) -> tuple[Node, Any]:
    value: Any | None = None
    match node.type:
        case "array":
            value = _handle_array_node(node)
        case "object":
            value = _handle_object_node(node)
        case "string":
            value = node, node.text[1:-1].decode("utf-8")
        case "number":
            try:
                value = node, int(node.text.decode("utf-8"))
            except ValueError:
                value = node, float(node.text.decode("utf-8"))
        case "true":
            value = node, True
        case "false":
            value = node, False
        case "null":
            value = node, None
        case _:
            raise UnexpectedNode(node)
    return value


def parse_json_with_tree_sitter(json: str) -> IndexedDict | IndexedList:
    parser = Parser()
    parser.set_language(
        Language(
            os.path.join(
                os.environ["TREE_SITTETR_PARSERS_DIR"],
                "json.so",
            ),
            "json",
        )
    )

    result = parser.parse(json.encode("utf-8"))
    value: IndexedDict | IndexedList | None = None
    for child in result.root_node.children:
        try:
            _, value = handle_json_node(child)
        except UnexpectedNode:
            LOGGER.exception(
                "Unexpected node",
                extra={
                    "extra": {
                        "result": result,
                        "node": child,
                        "json": json,
                    },
                },
            )
            break
    if not value:
        raise ValueError("Could not parse json")

    return value



from fluid_sbom import (
    advisories,
)
from fluid_sbom.file.location import (
    Location,
)
from fluid_sbom.file.resolver import (
    Resolver,
)
from fluid_sbom.internal.package_information.python import (
    get_pypi_package,
)
from fluid_sbom.pkg.cataloger.python.parse_wheel_egg_metadata import (
    ParsedData,
)
from fluid_sbom.pkg.language import (
    Language,
)
from fluid_sbom.pkg.package import (
    Artifact,
    HealthMetadata,
    Package,
)
from fluid_sbom.pkg.python import (
    PythonPackage,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
from fluid_sbom.utils.exceptions import (
    UnexpectedValueType,
)
from fluid_sbom.utils.file import (
    Digest,
)
import logging
from packageurl import (
    PackageURL,
)
from pydantic import (
    ValidationError,
)
from typing import (
    Any,
)

LOGGER = logging.getLogger(__name__)


def new_package_for_package(
    _resolver: Resolver, data: ParsedData, sources: Location
) -> Package | None:
    try:
        return Package(
            name=data.python_package.name,
            version=data.python_package.version or "",
            p_url=package_url(
                data.python_package.name,
                data.python_package.version or "",
                data.python_package,
            ),
            locations=[sources],
            language=Language.PYTHON,
            type=PackageType.PythonPkg,
            metadata=data.python_package,
            licenses=[data.licenses],
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types "
            "are incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": sources.path(),
                }
            },
        )
        return None


def package_url(name: str, version: str, package: PythonPackage | None) -> str:
    return PackageURL(
        type="pypi",
        namespace="",
        name=name,
        version=version,
        qualifiers=_purl_qualifiers_for_package(package),
        subpath="",
    ).to_string()


def _purl_qualifiers_for_package(
    package: PythonPackage | None,
) -> dict[str, str]:
    if not package:
        return {}
    if package.direct_url_origin and package.direct_url_origin.vcs:
        url = package.direct_url_origin
        return {"vcs_url": f"{url.vcs}+{url.url}@{url.commit_id}"}
    return {}


def _set_health_metadata_artifact(
    *,
    package: Package,
    current_package: dict[str, Any] | None,
) -> Package:
    if not current_package or not package.health_metadata:
        return package

    if url := next(
        (x for x in current_package["urls"] if x["url"].endswith(".tar.gz")),
        None,
    ):
        digest_value = url.get("digests", {}).get("sha256") or None
        package.health_metadata.artifact = Artifact(
            url=url.get("url"),
            integrity=Digest(
                algorithm="sha256" if digest_value else None,
                value=digest_value,
            ),
        )
    else:
        package.health_metadata.artifact = Artifact(
            url=f"https://pypi.org/pypi/{package.name}",
        )
    return package


def _set_authors(
    *,
    package: Package,
    pypi_package: dict[str, Any],
) -> Package:
    package_info = pypi_package["info"]
    if package.health_metadata and "author" in package_info:
        author: str | None = None
        package_author = package_info["author"]
        author_email = package_info.get("author_email")
        if isinstance(package_author, str) and package_author:
            author = package_author
        if not author and author_email:
            author = author_email
        if author and author_email and author_email not in author:
            author = f"{author} <{author_email}>"
        package.health_metadata.authors = author
    return package


def complete_package(package: Package) -> Package:
    pkg_advisories = advisories.get_package_advisories(package)
    if pkg_advisories:
        package.advisories = pkg_advisories
    current_package = get_pypi_package(package.name, package.version)
    pypi_package = get_pypi_package(package.name)
    if not pypi_package:
        return package
    try:
        package.health_metadata = HealthMetadata(
            latest_version=pypi_package["info"]["version"],
            latest_version_created_at=(
                pypi_package["releases"][pypi_package["info"]["version"]][0][
                    "upload_time_iso_8601"
                ]
            ),
        )
    except IndexError:
        return package
    package = _set_health_metadata_artifact(
        package=package, current_package=current_package
    )
    package = _set_authors(package=package, pypi_package=pypi_package)
    if not package.licenses and (
        licenses := pypi_package["info"].get("license")
    ):
        if isinstance(licenses, str):
            package.licenses = [licenses]
        else:
            raise UnexpectedValueType("license must be an string")
    return package



from fluid_sbom.pkg.cataloger.alpine import (
    package as package_alpine,
)
from fluid_sbom.pkg.cataloger.dart import (
    package as package_dart,
)
from fluid_sbom.pkg.cataloger.debian import (
    package as package_debian,
)
from fluid_sbom.pkg.cataloger.dotnet import (
    package as package_dotnet,
)
from fluid_sbom.pkg.cataloger.java import (
    package as package_java,
)
from fluid_sbom.pkg.cataloger.javascript import (
    package as package_js,
)
from fluid_sbom.pkg.cataloger.php import (
    package as package_php,
)
from fluid_sbom.pkg.cataloger.python import (
    package as package_python,
)
from fluid_sbom.pkg.cataloger.ruby import (
    package as package_ruby,
)
from fluid_sbom.pkg.cataloger.rust import (
    package as package_rust,
)
from fluid_sbom.pkg.package import (
    Package,
)
from fluid_sbom.pkg.type import (
    PackageType,
)
import logging
from pydantic import (
    ValidationError,
)

LOGGER = logging.getLogger(__name__)


def complete_package(package: Package) -> Package | None:
    completion_map = {
        PackageType.NpmPkg: package_js.complete_package,
        PackageType.DartPubPkg: package_dart.complete_package,
        PackageType.DotnetPkg: package_dotnet.complete_package,
        PackageType.JavaPkg: package_java.complete_package,
        PackageType.PhpComposerPkg: package_php.complete_package,
        PackageType.PythonPkg: package_python.complete_package,
        PackageType.GemPkg: package_ruby.complete_package,
        PackageType.RustPkg: package_rust.complete_package,
        PackageType.DebPkg: package_debian.complete_package,
        PackageType.ApkPkg: package_alpine.complete_package,
    }

    try:
        if package.type in completion_map:
            package = completion_map[package.type](package)  # type: ignore
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are "
            "incorrect.",
            extra={
                "extra": {
                    "exception": ex.errors(include_url=False),
                    "location": package.locations,
                }
            },
        )
        return None

    return package



from fluid_sbom.pkg.cataloger.generic.cataloger import (
    Request,
)
from fluid_sbom.pkg.cataloger.java.archive_parser import (
    parse_java_archive,
)
from fluid_sbom.pkg.cataloger.java.parse_android_apk import (
    parse_apk,
)
from fluid_sbom.pkg.cataloger.java.parse_gradle_lockfile import (
    parse_gradle_lockfile,
)
from fluid_sbom.pkg.cataloger.java.parse_pom_xml import (
    parse_pom_xml,
)
from fnmatch import (
    fnmatch,
)
import reactivex
from reactivex.abc import (
    ObserverBase,
    SchedulerBase,
)


def on_next_java(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if any(
                    fnmatch(value, x)
                    for x in (
                        "**/pom.xml",
                        "/pom.xml",
                        "pom.xml",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_pom_xml,
                            parser_name="java-parse-pom-xml",
                        )
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        "**/*.jar",
                        "*.jar",
                        "**/*.war",
                        "*.war",
                        "**/*.ear",
                        "*.ear",
                        "**/*.par",
                        "*.par",
                        "**/*.sar",
                        "*.sar",
                        "**/*.nar",
                        "*.nar",
                        "**/*.jpi",
                        "*.jpi",
                        "**/*.hpi",
                        "*.hpi",
                        "**/*.lpkg",
                        "*.lpkg",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_java_archive,
                            parser_name="java-archive-parse",
                        )
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        "gradle.lockfile*",
                        "**/gradle.lockfile*",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_gradle_lockfile,
                            parser_name="java-parse-gradle-lock",
                        )
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        "**/*.apk",
                        "*.apk",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_apk,
                            parser_name="java-parse-apk",
                        )
                    )
            except Exception as ex:  # pylint:disable=broad-exception-caught
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)
